/*
 * Try to generate a fairly good random value. We gather all sorts of system
 * status and put it through a hash algorithm.
 *
 * This is a weak imitation of Ted Tso's /dev/random that can be used
 * if running on kernels earlier than 1.3.3x (where /dev/random was
 * introduced).
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */
#include <sys/time.h>
#include <unistd.h>
#include <time.h>

#include "misc.h"
#include "sha.h"

static struct infile {
	char *		fname;
	unsigned int	maxvalue,
			curvalue,
			incvalue;
} sources[] = {
	{ "/proc/meminfo",	500,	500,	100 },
	{ "/proc/stat",		500,	500,	200 }, 
	{ "/proc/loadavg",	200,	200,	20 },
	{ "/proc/uptime",	200,	200,	20 },
	{ "/proc/kcore",	10000,	10000,	10000 }, 
	{ "/proc/net/dev",	200,	200,	100 }, 
	{ "/proc/net/arp",	200,	200,	10 }, 
	{ "/proc/net/snmp",	100,	100,	50 }, 
	{ "/proc/net/tcp",	100,	100,	20 }, 
	{ "/proc/net/udp",	100,	100,	20 }, 
	{ "/proc/net/unix",	100,	100,	10 }, 
	{ "/proc/net/raw",	100,	100,	10 }, 
	{ NULL,			0,	0,	0 }
};
#define MAXSRCS		(sizeof(sources)/sizeof(sources[0]) - 1)

#define MINRANDOMNESS	1500

static unsigned char	lastkey[SHA_DIGESTSIZE];
static time_t		lasttime = 0;

static int		eatfile(SHA_INFO *ctxt, char *fname);

int
weakrandomkey(unsigned char *keyout, int len)
{
	struct infile *	src;
	struct SHA_INFO	ctxt;
	struct timeval	nowtv;
	time_t		now, delta;
	int		entropy = 0, inc, i;

	gettimeofday(&nowtv, NULL);

	sha_init(&ctxt);
	sha_update(&ctxt, lastkey, SHA_DIGESTSIZE);
	sha_update(&ctxt, (unsigned char *) &nowtv, sizeof(nowtv));

	now = time(NULL);

again:
	if (lasttime) {
		delta = now - lasttime;
		for (i = 0, src = sources; i < MAXSRCS; i++, src++) {
			inc = delta * src->incvalue;
			if ((src->curvalue += inc) > src->maxvalue)
				src->curvalue = src->maxvalue;
		}
	}
	lasttime = now;

	for (i = 0, src = sources; i < MAXSRCS; i++, src++) {
		if (!src->curvalue || !eatfile(&ctxt, src->fname))
			continue;
		entropy += src->curvalue;
		src->curvalue >>= 1;
		if (entropy >= MINRANDOMNESS) {
			sha_final(&ctxt);
			if (len > SHA_DIGESTSIZE)
				len = SHA_DIGESTSIZE;
			memcpy(keyout, ctxt.digest, len);
			memcpy(lastkey, ctxt.digest, SHA_DIGESTSIZE);
			return len;
		}
	}

	/* Sigh. Not enough information; sleep a second. */
	do {
		sleep(1);
		now = time(NULL);
	} while (now == lasttime);

	goto again;
}

static int
eatfile(SHA_INFO *ctxt, char *fname)
{
	char	buffer[128];
	FILE	*fp;
	int	len;

	if (!(fp = fopen(fname, "r")))
		return 0;
	while ((len = fread(buffer, 1, sizeof(buffer), fp)) != 0)
		sha_update(ctxt, buffer, len);
	fclose(fp);
	return 1;
}
