/*
 * genkeys	Generate keys for NFS file handle signatures.
 *		Parts of this may have to go into libnfs so that
 *		mountd will be able to generate keys for netgroup and
 *		subnet hosts on the fly.
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */

#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <ctype.h>
#include <fcntl.h>
#include <nfslib.h>
#include <misc.h>
#include <xio.h>

static int	opt_a = 0,
		opt_d = 0,
		opt_v = 0;

typedef struct keyinfo {
	char	*name;
	int	type;
	int	len;
} keyinfo_t;

static char *		keyfile = _PATH_NFSKEYS;
static keyinfo_t	keyinfo[] = {
    {	"null",	CLE_KEY_NULL,	0	},
    {	"md5",	CLE_KEY_MD5,	16	},
    {	"sha",	CLE_KEY_SHA,	20	},
    {	"none",	CLE_KEY_NONE,	0	},
    {	NULL,	-1,		0	},
};


static void	usage(int excode);
static void	dumpkeys(void);
static int	getnfskeytype(char *type);
static int	getnfskeylen(int type);
static char *	getnfskeyname(int type);
static int	genkey(struct nfskey *kp, int len);
static int	putkey(struct nfskeyent *kep);
static int	string_to_key(char *skey, struct nfskey *kp);
static char *	key_to_string(struct nfskey *kp, char *skey);

int
main(int argc, char **argv)
{
	struct nfskeyent	ke;
	char	*hostname = NULL;
	char	*skey = NULL;
	char	*sktype = NULL;
	char	c;

	if (getuid() != geteuid() || getgid() != getegid()) {
		fprintf(stderr,
			"You should not make this program setuid or setgid.\n"
			"Refusing to run, sorry.\n");
		exit(1);
	}

	while ((c = getopt(argc, argv, "adf:h:k:t:v")) != EOF) {
		switch (c) {
		case 'a':
			opt_a++;
			break;
		case 'd':
			opt_d++;
			break;
		case 'f':
			keyfile = optarg;
			break;
		case 'h':
			hostname = optarg;
			break;
		case 'k':
			skey = optarg;
			break;
		case 't':
			sktype = optarg;
			break;
		case 'v':
			opt_v++;
			break;
		default:
			usage(c != '?');
		}
	}

	if (!opt_a && !opt_d) {
		dumpkeys();
		exit (0);
	}
	if (opt_d && opt_a) {
		fprintf(stderr, "only one of -a and -d allowed\n");
		exit (1);
	}
	memset(&ke, 0, sizeof(ke));
	if (hostname) {
		struct hostent *hp;

		if (!(hp = gethostbyname(hostname))) {
			fprintf(stderr, "can't resolve hostname %s\n",
				hostname);
			exit (1);
		}
		if (strlen(hp->h_name) >= NFSCLNT_IDMAX) {
			fprintf(stderr, "hostname too long!\n");
			exit (1);
		}
		strcpy(ke.k_hostname, hp->h_name);
	} else {
		fprintf(stderr, "no hostname specified\n");
		exit (1);
	}
	if (sktype) {
		ke.k_key.k_type = getnfskeytype(sktype);
		if (ke.k_key.k_type == CLE_KEY_NONE) {
			fprintf(stderr, "unknown key type %s\n",
				sktype);
			exit (1);
		}
	} else {
		ke.k_key.k_type = CLE_KEY_NULL;
	}
	if (skey) {
		if (!string_to_key(skey, &ke.k_key))
			exit (1);
	} else {
		ke.k_key.k_len = 0;
	}
	ke.k_resvd = 0;
	if (opt_a && !ke.k_key.k_len) {
		int	len = getnfskeylen(ke.k_key.k_type);

		if (len && !genkey(&ke.k_key, len))
			exit (1);
	} else if (opt_d) {
		ke.k_key.k_type = CLE_KEY_NONE;
	}
	if (!putkey(&ke))
		return 1;

	return 0;
}

static void
usage(int excode)
{
	fprintf(stderr, "usage: nfsgenkey [-adv] [-h hostname] [-t type]\n");
	exit (excode);
}

static void
dumpkeys(void)
{
	struct nfskeyent * kep;
	char		buf[2 * sizeof(kep->k_key.k_key) + 1];

	if (!setnfskeyent(keyfile)) {
		fprintf(stderr, "nfsgenkeys: can't open %s: %s\n",
				keyfile, strerror(errno));
		return;
	}
	while ((kep = getnfskeyent()) != NULL) {
		if (kep->k_key.k_type == CLE_KEY_NONE)
			continue;
		printf("%-6s ", getnfskeyname(kep->k_key.k_type));
		if (opt_v) {
			printf("%-40s ", 
				key_to_string(&kep->k_key, buf));
		}
		printf("%s\n", kep->k_hostname);
	}
	endnfskeyent();
}

static int
getnfskeytype(char *type)
{
	keyinfo_t	*kip;

	for (kip = keyinfo; kip->name; kip++)
		if (!strcasecmp(kip->name, type))
			return kip->type;
	return CLE_KEY_NONE;
}

static int
getnfskeylen(int type)
{
	keyinfo_t	*kip;

	for (kip = keyinfo; kip->name; kip++)
		if (kip->type == type)
			return kip->len;
	return 0;
}

static char *
getnfskeyname(int type)
{
	keyinfo_t	*kip;

	for (kip = keyinfo; kip->name; kip++)
		if (kip->type == type)
			return kip->name;
	return "???";
}

static int
genkey(struct nfskey *kp, int len)
{
	int	klen;

	if (!(klen = randomkey((unsigned char *)&kp->k_key, len))) {
		fprintf(stderr, "warning: using weak random "
				"number generation\n");
		if (!getuid() || !geteuid())
			printf("Please wait...\n");
		klen = weakrandomkey((unsigned char *)&kp->k_key, len);
		if (!klen) {
			fprintf(stderr, "couldn't generate key\n");
			return 0;
		}
	}
	if (klen < len) {
		fprintf(stderr, "generated key too small: %d\n", klen);
		return 0;
	}
	kp->k_len = len;
	return len;
}

static int
putkey(struct nfskeyent *kep)
{
	struct nfskeyent *rkep;
	int	len = getnfskeylen(kep->k_key.k_type);
	int	lockid, fd;
	off_t	fpos = 0, floc = 1;
	FILE	*fp;

	if (!kep->k_key.k_len && len && !genkey(&kep->k_key, len))
		return 0;

	do {
		lockid = xflock(keyfile, "r+");
		if (lockid < 0) {
			if ((fd = open(keyfile, O_CREAT|O_EXCL, 0600)) < 0)
				break;
			close(fd);
		}
	} while (lockid < 0);

	if (lockid < 0) {
		fprintf(stderr, "could not lock %s\n", keyfile);
		return 0;
	}
	if (!(fp = fsetnfskeyent(keyfile, "r+"))) {
		fprintf(stderr, "can't open %s for update\n", keyfile);
		xfunlock(lockid);
		return 0;
	}
	while ((rkep = fgetnfskeyent(fp)) != NULL) {
		if (!strcmp(rkep->k_hostname, kep->k_hostname)) {
			floc = fpos;
			break;
		}
		if (rkep->k_key.k_type == CLE_KEY_NONE)
			floc = fpos;
		fpos = ftell(fp);
	}
	if (floc != 1)
		fseek(fp, floc, SEEK_SET);
	fputnfskeyent(fp, kep);
	fendnfskeyent(fp);
	xfunlock(lockid);

	return 0;
}

static int
xtod(char c)
{
	if ('0' <= c && c <= '9')
		return c - '0';
	if ('a' <= c && c <= 'f')
		return c - 'a' + 10;
	return -1;
}


static int
string_to_key(char *skey, struct nfskey *kp)
{
	char	*sp = skey, c0, c1;
	int	i = 0;

	while ((c0 = *sp++) != '\0') {
		if ((c1 = *sp++) == '\0') {
			fprintf(stderr,
				"odd number of digits in key!\n");
			return 0;
		}
		if ((c0 = xtod(tolower(c0))) < 0 ||
		    (c1 = xtod(tolower(c1))) < 0) {
			fprintf(stderr,
				"invalid digit in key!\n");
			return 0;
		}
		kp->k_key[i++] = (c0 << 4) | c1;
	}
	kp->k_len = i;
	return 1;
}

static char *
key_to_string(struct nfskey *kp, char *skey)
{
	int	i;

	skey[0] = '\0';
	for (i = 0; i < kp->k_len; i++) {
		sprintf(skey + 2*i, "%02x", kp->k_key[i]);
	}
	return skey;
}
