/*
 * auth.c		nfsd-specific RPC credential stuff.
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#ifdef MODULE
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/param.h>
#include <linux/sched.h>
#include <linux/string.h>
#include "nfsd.h"

static int	authnull_decode(svc_auth *ahp);
static int	authnull_setuser(svc_req *rqstp, struct svc_export *exp);
static int	authunix_decode(svc_auth *ahp);
static int	authunix_setuser(svc_req *rqstp, struct svc_export *exp);

static svc_authops	authnull_ops = {
	authnull_decode,
	NULL,
	authnull_setuser,
	NULL,
	NULL,
};

static svc_authops	authunix_ops = {
	authunix_decode,
	NULL,
	authunix_setuser,
	NULL,
	NULL,
};

static svc_authops *	auth_ops[AUTH_MAXFLAVOR] = {
	&authnull_ops,			/* NULL */
	&authunix_ops,			/* UNIX */
	NULL,				/* SHORT */
	NULL,				/* DES */
};

/*
 * Get the auth operations given the auth flavor.
 */
svc_authops *
auth_getops(u32 flavor)
{
	if (flavor < sizeof(auth_ops)/sizeof(auth_ops[0]))
		return auth_ops[flavor];
	return NULL;
}

static int
authnull_decode(svc_auth *ahp)
{
	return (ahp->ah_crflavor == 0) && (ahp->ah_crlen == 0)
	    && (ahp->ah_vfflavor == 0) && (ahp->ah_vflen == 0);
}

static int
authnull_setuser(svc_req *rqstp, svc_export *exp)
{
	current->fsuid = exp->ex_anon_uid;
	current->fsgid = exp->ex_anon_gid;
	current->groups[0] = NOGROUP;
	return 1;
}

/*
 * Decode unix credentials. We ignore the machine name, because it's
 * pretty worthless anyway. This also saves us stupid malloc's and
 * free's.
 */
static int
authunix_decode(svc_auth *ahp)
{
	struct authunix_parms	*cred;
	u32 *	buf = ahp->ah_crbase;
	int	len = ahp->ah_crlen;
	u32	stamp, l, uid, gid, glen;

	/* Require NULL verifier */
	if (ahp->ah_vfflavor != 0 || ahp->ah_vflen != 0)
		return 0;

	if (len-- < 1)
		return 0;
	stamp = ntohl(*buf++);		/* time stamp */

	/* Skip the machine name. */
	if (len-- < 1)
		return 0;
	l = ntohl(*buf++);
	if (l > 255 || (l = ((l + 3) >> 2)) > len)
		return 0;
	buf += l;
	len -= l;

	if ((len -= 3) < 0)
		return 0;
	uid = ntohl(*buf++);		/* uid */
	gid = ntohl(*buf++);		/* gid */
	glen = ntohl(*buf++);		/* suppl. gid count */

	if (glen > len || glen > NGRPS || glen > NGROUPS)
		return 0;

	/* Copy credentials in-place */
	cred = (struct authunix_parms *)(ahp->ah_crbase);
	cred->aup_stamp = stamp;
	cred->aup_uid = uid;
	cred->aup_gid = gid;
	cred->aup_len = glen;
	for (l = 0; l < glen; l++)
		cred->aup_gids[l] = ntohl(*buf++);

	return 1;
}

static int
authunix_setuser(svc_req *rqstp, svc_export *exp)
{
	struct authunix_parms	*cred;
	int			i;

	cred = (struct authunix_parms *)(rqstp->rq_cred.ah_crbase);

	current->fsuid = auth_luid(exp, cred->aup_uid);
	current->fsgid = auth_lgid(exp, cred->aup_gid);
	if (cred->aup_len > NGROUPS) {
		printk(KERN_WARNING "nfsd: aup_len too large (%d)!\n",
				(int) cred->aup_len);
		cred->aup_len = 0;
	}
	for (i = 0; i < cred->aup_len; i++)
		current->groups[i] = auth_lgid(exp, cred->aup_gids[i]);
	if (i < NGROUPS)
		current->groups[i] = NOGROUP;
	return 1;
}

/*
 * The following deals with the communication between the kernel and
 * NFS auth daemons (to be used by authentication flavors like Kerberos
 * or DES).
 */
static struct authd_state *	authd_state[AUTH_MAXFLAVOR];
static unsigned long		authd_serial = NFSAUTHD_NO_SERIAL;

int
authd_call(int flavor, void *argdata, int arglen, void *resdata, int reslen)
{
	struct authd_state	*authd;
	struct auth_args	*argp;
	int			serial;
	int			status;

	if (!(authd = authd_state[flavor]))
		return -EACCES;

	/* Drop request if someone else is calling authd */
	if (authd->ad_inprog || authd->ad_dead > jiffies)
		return -EACCES;
	authd->ad_inprog++;

	current->timeout = NFSAUTHD_TIMEOUT;
	if (++authd_serial == NFSAUTHD_NO_SERIAL)
		authd_serial++;
	serial = authd_serial;
	status = -EACCES;

	while ((argp = authd->ad_argp) != NULL) {
		/* Copy our arguments to the daemon's result struct */
		argp->ac_serial = serial;
		argp->ac_length = arglen;
		memcpy(argp->ac_data, argdata, arglen);
		wake_up(&authd->ad_server);

		/* Go to sleep until the daemon wakes us up again */
		interruptible_sleep_on(&authd->ad_client);
		if (!current->timeout) {
			/* Ooops, daemon fell over */
			authd->ad_dead = jiffies + NFSAUTHD_BACKOFF;
			break;
		}

		/* Copy daemon's argument struct to our result */
		if ((argp = authd->ad_resp) && argp->ac_serial == serial) {
			if (argp->ac_length > reslen)
				break;
			memcpy(resdata, argp->ac_data, reslen);
			status = argp->ac_status;
			break;
		}
	}
	current->timeout = 0;
	authd->ad_inprog = 0;

	return status;
}

int
authd_stub(struct auth_args *argp, struct auth_args *resp)
{
	struct authd_state	*authd;

	if (!suser())
		return -EPERM;
	if (argp->ac_flavor >= AUTH_MAXFLAVOR)
		return -EINVAL;
	if (!(authd = authd_state[argp->ac_flavor]))
		return -EINVAL;
	if (authd->ad_running)
		return -EBUSY;

	authd->ad_argp = argp;
	authd->ad_resp = resp;

	wake_up(&authd->ad_client);
	interruptible_sleep_on(&authd->ad_server);

	authd->ad_argp = NULL;
	authd->ad_resp = NULL;

	return 0;
}
