/*
 * auth.h		nfsd-specific authentication stuff.
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#ifndef __NFSD_AUTH_H__
#define __NFSD_AUTH_H__

/*
 * Various RPC authentication flavors.
 */
#define AUTH_NULL	0
#define AUTH_UNIX	1
#define AUTH_SHORT	2
#define AUTH_DES	3
#define AUTH_MAXFLAVOR	8

/*
 * This is the argument/result struct for the authentication daemon(s).
 */
#define NFSAUTHD_MAXDATA	500
struct auth_args {
	int		ac_flavor;
	int		ac_serial;
	int		ac_length;
	int		ac_status;
	char		ac_data[NFSAUTHD_MAXDATA];
};


#ifdef __KERNEL__

typedef struct svc_authops	svc_authops;

typedef struct svc_auth {
	svc_authops *	ah_op;
	u32		ah_crflavor;
	void *		ah_crbase;
	int		ah_crlen;
	u32		ah_vfflavor;
	void *		ah_vfbase;
	int		ah_vflen;
} svc_auth;

struct svc_buf;		/* forward decl */
struct svc_req;

/*
 * Standard operations for RPC authentication schemes.
 */
struct svc_authops {
	/* Decode authinfo */
	int		(*decode)(svc_auth *ahp);
	void		(*free)(svc_auth *ahp);

	/* Set fsuid/fsgid and supplemental groups */
	int		(*setuser)(struct svc_req *rqstp, svc_export *exp);

	/* Check client credential or verifier */
	int		(*validate)(struct svc_req *rqstp, u32 *errp);

	/* Add verifier to response. */
	int		(*addverf)(struct svc_req *rqstp,
						struct svc_buf *resp,
						u32 *errp);
};

/*
 * Decoded AUTH_UNIX data. This is different from what's in the RPC lib.
 */
#define NGRPS		16
struct authunix_parms {
	u32		aup_stamp;
	u32		aup_uid;
	u32		aup_gid;
	u32		aup_len;
	u32		aup_gids[NGRPS];
};

/*
 * State information for authentication daemon.
 */
#define NFSAUTHD_NO_SERIAL	0
#define NFSAUTHD_TIMEOUT	(60 * HZ)
#define NFSAUTHD_BACKOFF	(5 * NFSAUTHD_TIMEOUT)

struct authd_state {
	struct wait_queue *	ad_client;
	struct wait_queue *	ad_server;
	struct auth_args *	ad_argp;
	struct auth_args *	ad_resp;
	unsigned long		ad_dead;
	unsigned char		ad_inprog;
	unsigned char		ad_running;
};


svc_authops *		auth_getops(u32 flavor);
int			authd_call(int flavor, void *argdata, int arglen,
					       void *resdata, int reslen);
int			authd_stub(struct auth_args *, struct auth_args *);

/*
 * Inline functions for uid/gid mapping. Should probably be moved to
 * export.h
 */
static inline uid_t	auth_ruid(svc_export *exp, uid_t uid)
{
	uid_t	*uip;

	if (!(uip = exp->ex_client->cl_umap.um_luid[uid >> UGID_SHIFT]))
		return exp->ex_anon_uid;
	return uip[uid & UGID_MASK];
}


static inline gid_t	auth_rgid(svc_export *exp, gid_t gid)
{
	gid_t	*gip;

	if (!(gip = exp->ex_client->cl_umap.um_lgid[gid >> UGID_SHIFT]))
		return exp->ex_anon_gid;
	return gip[gid & UGID_MASK];
}

static inline uid_t	auth_luid(svc_export *exp, uid_t uid)
{
	uid_t	*uip;

	if ((exp->ex_flags & NFSEXP_ALLSQUASH) ||
	    ((exp->ex_flags & NFSEXP_ROOTSQUASH) && !uid))
		return exp->ex_anon_uid;
	if (!(exp->ex_flags & NFSEXP_UIDMAP))
		return uid;

	if ((uid >> UGID_SHIFT) > UGID_NRENTRIES) {
		printk(KERN_WARNING "nfsd: bad luid umap size (uid = %d)!\n",
			uid);
		return exp->ex_anon_uid;
	}
	if (!(uip = exp->ex_client->cl_umap.um_luid[uid >> UGID_SHIFT]))
		return exp->ex_anon_uid;
	return uip[uid & UGID_MASK];
}

static inline gid_t	auth_lgid(svc_export *exp, gid_t gid)
{
	gid_t	*gip;

	if ((exp->ex_flags & NFSEXP_ALLSQUASH) ||
	    ((exp->ex_flags & NFSEXP_ROOTSQUASH) && !gid))
		return exp->ex_anon_gid;
	if (!(exp->ex_flags & NFSEXP_UIDMAP))
		return gid;

	if ((gid >> UGID_SHIFT) > UGID_NRENTRIES) {
		printk(KERN_WARNING "nfsd: bad lgid umap size (gid = %d)!\n",
			gid);
		return exp->ex_anon_gid;
	}
	if (!(gip = exp->ex_client->cl_umap.um_lgid[gid >> UGID_SHIFT]))
		return exp->ex_anon_gid;
	return gip[gid & UGID_MASK];
}

#endif /* __KERNEL__ */

#endif /* __NFSD_AUTH_H__ */
