/*
 * export.h		Public declarations for NFS exports. The definitions
 *			for the syscall interface are in nfsctl.h
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#ifndef _NFSEXPORT_H
#define _NFSEXPORT_H

#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/fs.h>

/*
 * Important limits for the exports stuff.
 */
#define NFSCLNT_IDMAX		1024
#define NFSCLNT_ADDRMAX		16
#define NFSCLNT_KEYMAX		32

/*
 * Export flags.
 */
#define NFSEXP_READONLY		0x0001
#define NFSEXP_INSECURE_PORT	0x0002
#define NFSEXP_ROOTSQUASH	0x0004
#define NFSEXP_ALLSQUASH	0x0008
#define NFSEXP_ASYNC		0x0010
#define NFSEXP_UIDMAP		0x0020
#define NFSEXP_SIGNFH		0x0040
#define NFSEXP_KERBEROS		0x0080		/* unused */
#define NFSEXP_SUNSECURE	0x0100
#define NFSEXP_CROSSMNT		0x0200		/* unused */
#define NFSEXP_ALLFLAGS		0x03FF


#ifdef __KERNEL__

/* The following are hashtable sizes and must be powers of 2 */
#define NFSCLNT_EXPMAX		16
#define NFSEXP_FHMAX		32

/* These must match the actual size of uid_t and gid_t */
#define UGID_BITS		16
#define UGID_SHIFT		8
#define UGID_MASK		((1 << UGID_SHIFT) - 1)
#define UGID_NRENTRIES		((1 << (UGID_BITS - UGID_SHIFT)) + 1)
#define UGID_NONE		((unsigned short)-1)

typedef struct svc_uidmap {
	uid_t			*um_ruid[UGID_NRENTRIES];
	uid_t			*um_luid[UGID_NRENTRIES];
	gid_t			*um_rgid[UGID_NRENTRIES];
	gid_t			*um_lgid[UGID_NRENTRIES];
} svc_uidmap;

typedef struct svc_client {
	struct svc_client *	cl_next;
	char			cl_ident[NFSCLNT_IDMAX];
	int			cl_idlen;
	int			cl_naddr;
	struct in_addr		cl_addr[NFSCLNT_ADDRMAX];
	struct svc_uidmap	cl_umap;
	struct svc_export *	cl_export[NFSCLNT_EXPMAX];
	int			cl_fhkeytype;
	int			cl_fhkeylen;
	unsigned char		cl_fhkey[NFSCLNT_KEYMAX];
	int			cl_count;
	int			cl_lock;
	struct wait_queue *	cl_wait;
} svc_client;

typedef struct svc_export {
	struct svc_export *	ex_next;
	char			ex_path[NFS_MAXPATHLEN+1];
	struct svc_export *	ex_parent;
	struct svc_client *	ex_client;
	int			ex_flags;
	struct inode *		ex_inode;
	dev_t			ex_dev;
	ino_t			ex_ino;
	uid_t			ex_anon_uid;
	gid_t			ex_anon_gid;
	struct svc_fh *		ex_fhcache[NFSEXP_FHMAX];
} svc_export;

#define EX_SECURE(exp)		(!((exp)->ex_flags & NFSEXP_INSECURE_PORT))
#define EX_ISSYNC(exp)		(!((exp)->ex_flags & NFSEXP_ASYNC))
#define EX_RDONLY(exp)		((exp)->ex_flags & NFSEXP_READONLY)
#define EX_CROSSMNT(exp)	((exp)->ex_flags & NFSEXP_CROSSMNT)

extern __inline__ int
exp_checkaddr(svc_client *clp, struct in_addr addr)
{
	struct in_addr	*ap = clp->cl_addr;
	int		i;

	for (i = clp->cl_naddr; i--; ap++)
		if (ap->s_addr == addr.s_addr)
			return 1;
	return 0;
}

#endif /* __KERNEL__ */

#endif /* _NFSEXPORT_H */

