/*
 * fileop.c		File operations used by nfsd. Some of these
 *			have been ripped from other parts of the kernel
 *			because they weren't in ksyms.c, others are
 *			partial duplicates with added or changed 
 *			functionality.
 *
 * Copyright (C) 1995,1996 Olaf Kirch <okir@monad.swb.de>
 */

#ifdef MODULE
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/linkage.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/ext2_fs.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <netinet/in.h>
#include "nfsd.h"

/* Symbol not exported */
static struct super_block *get_super(dev_t dev);

/* Missing external decls */
extern int	close_fp(struct file *filp);

/* Imported from fs/open.c */
#define ACC_MODE(x)	("\000\004\002\006"[(x) & O_ACCMODE])

/*
 * nfsd_lookup looks up one component of a pathname.
 */
int
nfsd_lookup(svc_req *rqstp, svc_fh *fhp, char *name, int len, svc_fh **result)
{
	struct svc_export	*exp = fhp->fh_export;
	struct super_block	*sb;
	struct inode		*dir, *inode;
	int			perm, err, dotdot = 0;

	*result = NULL;

	/* Obtain inode */
	if (!(dir = fh_inode(rqstp, fhp, S_IFDIR, MAY_NOP, &err)))
		return err;

	/* check permissions before traversing mount-points */
	perm = nfsd_permission(rqstp, exp, dir, MAY_EXEC);

	dotdot = (len == 2 && name[0] == '.' && name[1] == '.');
	if (dotdot) {
		if (dir == current->fs->root) {
			fhp->fh_count++;
			*result = fhp;
			return 0;
		}

		if (dir->i_dev == exp->ex_dev && dir->i_ino == exp->ex_ino) {
			fhp->fh_count++;
			*result = fhp;
			return 0;
		}
	} else if (len == 1 && name[0] == '.') {
		name[0] = 0;
		len = 0;
	} else if (dir->i_sb->s_magic == NFS_SUPER_MAGIC) {
		/* Client is trying to look up a name on an NFS mount */
		return nfserr_noent;
	}
	if (!dir->i_op || !dir->i_op->lookup)
		return nfserr_notdir;
 	if (perm != 0)
		return perm;
	if (!len) {
		fhp->fh_count++;
		*result = fhp;
		return 0;
	}

	dir->i_count++;		/* lookup eats the dir inode */
	err = dir->i_op->lookup(dir, name, len, &inode);
	if (err)
		return nfserrno(-err);

	/* Note that lookup() has already done a call to iget() so that
	 * the inode returned never refers to an inode covered by a mount.
	 * When this has happened, return the covered inode.
	 */
	if (!dotdot && (sb = inode->i_sb) && (inode == sb->s_mounted)) {
		iput(inode);
		inode = sb->s_covered;
		inode->i_count++;
	}

	if (!(fhp = fh_compose(rqstp, exp, inode))) {
		nfsd_iput(inode);
		return nfserr_io;	/* bad choice */
	}

	*result = fhp;
	return 0;
}

/*
 * Set various file attributes.
 */
int
nfsd_setattr(svc_req *rqstp, svc_fh *fhp, struct iattr *iap)
{
	struct inode	*inode;
	int		accmode = MAY_SATTR;
	int		ftype = 0;
	int		imode;
	int		err;

	if (iap->ia_valid & (ATTR_ATIME | ATTR_MTIME | ATTR_SIZE))
		accmode |= MAY_WRITE;
	if (iap->ia_valid & ATTR_SIZE)
		ftype = S_IFREG;

	/* nfsd_sattr_umap(fhp->fh_export, iap); */

	/* Get inode */
	if (!(inode = fh_inode(rqstp, fhp, ftype, accmode, &err)))
		return err;

	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		return nfserr_perm;

	/* No attribute changes on NFS mount points. */
	if (S_ISDIR(inode->i_mode) && nfsd_iscovered(inode))
		return nfserr_perm;

	/* The size case is special... */
	if (iap->ia_valid & ATTR_SIZE) {
		if ((err = get_write_access(inode)) != 0)
			return nfserrno(-err);
		inode->i_size = iap->ia_size;
		if (inode->i_op && inode->i_op->truncate)
			inode->i_op->truncate(inode);
		inode->i_dirt = 1;
		put_write_access(inode);
	}

	imode = inode->i_mode;
	if (iap->ia_valid & ATTR_MODE) {
		iap->ia_mode &= S_IALLUGO;
		imode = iap->ia_mode |= (imode & ~S_IALLUGO);
	}

	/* Revoke setuid/setgid bit on chown/chgrp */
	if ((iap->ia_valid & ATTR_UID) && (imode & S_ISUID)
	 && iap->ia_uid != inode->i_uid) {
		iap->ia_valid |= ATTR_MODE;
		iap->ia_mode = imode &= ~S_ISUID;
	}
	if ((iap->ia_valid & ATTR_GID) && (imode & S_ISGID)
	 && iap->ia_gid != inode->i_gid) {
		iap->ia_valid |= ATTR_MODE;
		iap->ia_mode = imode &= ~S_ISGID;
	}

	/* Change the attributes. The user creds for notify_change_ok
	 * have been set in fh_decode already.
	 */
	if (iap->ia_valid) {
		iap->ia_valid |= ATTR_CTIME;
		iap->ia_ctime = CURRENT_TIME;
		err = nfsd_notify_change(inode, iap);
		if (err)
			return nfserrno(-err);
		if (EX_ISSYNC(fhp->fh_export))
			write_inode(inode);
	}

	return 0;
}

/*
 * Open an existing file or directory. Permissions have already been
 * checked by fh_inode(), including if the file system is readonly.
 * The calling procedure should have made sure that the file is indeed
 * a regular file or directory.
 *
 * Partial copy from fs/open.c:do_open
 */
int
nfsd_open(svc_req *rqstp, svc_fh *fhp, int type, int flags, int omode,
					struct file **filpp)
{
	struct inode	*inode;
	struct file	*filp;
	int		mode, flag = flags;
	int		err;

	/* We're not interested in any other flags. (O_APPEND will
	 * only be used by the NFSv3 code).
	 */
	flags &= (O_ACCMODE | O_APPEND);

	mode = (flag + 1) & O_ACCMODE;
	if (mode)
		flag++;

	inode = fh_inode(rqstp, fhp, type,
			(mode&2)? MAY_WRITE : MAY_READ, &err);
	if (!inode)
		return err;

	if (IS_APPEND(inode) && ((flag & 2) && !(flag & O_APPEND)))
		return nfserr_perm;

	if (mode & 2) {
		err = get_write_access(inode);
		if (err)
			return nfserrno(-err);
		if (EX_ISSYNC(fhp->fh_export))
			flags |= O_SYNC;
	}

	if (!(filp = get_empty_filp())) {
		if (mode & 2)
			put_write_access(inode);
		return nfserr_io;
	}

	filp->f_count = 1;
	filp->f_flags = flags;
	filp->f_mode = mode;
	filp->f_inode = inode;
	filp->f_pos = 0;
	filp->f_reada = 0;
	if (inode->i_op)
		filp->f_op = inode->i_op->default_file_ops;
	else
		filp->f_op = NULL;
	if (filp->f_op && filp->f_op->open) {
		err = filp->f_op->open(inode, filp);
		if (err) {
			if (mode & 2)
				put_write_access(inode);
			filp->f_count--;
			return nfserrno(-err);
		}
	}

	inode->i_count++;
	*filpp = filp;
	return 0;
}

/*
 * Close a file.
 */
void
nfsd_close(struct file *filp)
{
#if 0
	/* This is for the case where the nfs_proc routines pass in a
	 * file struct that's on their stack.
	 */
	struct inode	*inode;

	inode = filp->f_inode;
	if (!inode->i_count)
		printk(KERN_WARNING "nfsd: inode count == 0!\n");
	if (filp->f_op && filp->f_op->release)
		filp->f_op->release(inode, filp);
	if (filp->f_mode & 2)
		put_write_access(inode);
	nfsd_iput(inode);
#else
	close_fp(filp);
#endif
}

/*
 * Read data from a file. count must contain the requested read count
 * on entry. On return, *count contains the number of bytes actually read.
 */
int
nfsd_read(svc_req *rqstp, svc_fh *fhp, loff_t offset, char *buf,
						unsigned long *count)
{
	struct file	*filp;
	unsigned long	oldfs;
	int		err;

	if ((err = nfsd_open(rqstp, fhp, S_IFREG, O_RDONLY, 0, &filp)) != 0)
		return err;
	if (!filp->f_op || !filp->f_op->read) {
		nfsd_close(filp);
		return nfserr_perm;
	}

	filp->f_reada = 1;	/* currently, we always force a readahead */
	filp->f_pos = offset;

#if 1
	oldfs = get_fs();
	set_fs(get_ds());
	err = filp->f_op->read(filp->f_inode, filp, buf, *count);
	set_fs(oldfs);
#else
	/* Test code */
	memset(buf, 0, *count);
	err = *count;
#endif

	nfsd_close(filp);

	if (err < 0)
		return nfserrno(-err);
	*count = err;
	return 0;
}

/*
 * Write data to a file.
 */
int
nfsd_write(svc_req *rqstp, svc_fh *fhp, loff_t offset, char *buf,
					unsigned long cnt)
{
	struct file *	filp;
	unsigned long	oldfs;
	int		err;

	if (!cnt)
		return 0;
	if ((err = nfsd_open(rqstp, fhp, S_IFREG, O_WRONLY, 0, &filp)) != 0)
		return err;
	if (!filp->f_op || !filp->f_op->write) {
		nfsd_close(filp);
		return nfserr_perm;
	}

	filp->f_pos = offset;

	oldfs = get_fs();
	set_fs(get_ds());
	err = filp->f_op->write(filp->f_inode, filp, buf, cnt);
	set_fs(oldfs);

	if (!err && filp->f_inode->i_dirt && EX_ISSYNC(fhp->fh_export))
		write_inode(filp->f_inode);

	nfsd_close(filp);

	return (err < 0)? nfserrno(-err) : 0;
}

/*
 * Create a regular file.
 */
int
nfsd_create(svc_req *rqstp, svc_fh *fhp, char *fname, int flen, int mode,
						svc_fh **result)
{
	struct inode *dir, *inode;
	int	err;

	if (!flen)
		return nfserr_perm;
	if (!(dir = fh_inode(rqstp, fhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err)))
		return err;
	if (nfsd_iscovered(dir))
		return nfserr_acces;
	if (!dir->i_op || !dir->i_op->create)
		return nfserr_perm;

	dir->i_count++;
	down(&dir->i_sem);
	err = dir->i_op->create(dir, fname, flen, mode, &inode);
	up(&dir->i_sem);

	*result = fh_compose(rqstp, fhp->fh_export, inode);

	return err? nfserrno(-err) : 0;
}

/*
 * Create a special file.
 */
int
nfsd_mknod(struct inode *dir, char *fname, int flen, int mode, dev_t rdev)
{
	int	err;

	if (!dir->i_op || !dir->i_op->mknod) {
		return nfserr_perm;
	}

	dir->i_count++;
	down(&dir->i_sem);
	err = dir->i_op->mknod(dir, fname, flen, mode, rdev);
	up(&dir->i_sem);

	return err? nfserrno(-err) : 0;
}

/*
 * Truncate a file. The calling routine is required to check for immutables
 * and append-only itself. It must also make sure to update the ctime
 * field and call notify_change.
 */
int
nfsd_truncate(svc_req *rqstp, struct inode *inode, unsigned long size)
{
	int	err;

	if ((err = get_write_access(inode)) != 0)
		return nfserrno(-err);
	inode->i_size = size;
	if (inode->i_op && inode->i_op->truncate)
		inode->i_op->truncate(inode);
	inode->i_dirt = 1;
	put_write_access(inode);

	return 0;
}

/*
 * Read a symlink. On entry, *lenp must contain the maximum path length that
 * fits into the buffer. On return, it contains the true length.
 */
int
nfsd_readlink(svc_req *rqstp, svc_fh *fhp, char *buf, int *lenp)
{
	struct inode	*inode;
	unsigned long	oldfs;
	int		err;

	if (!(inode = fh_inode(rqstp, fhp, S_IFLNK, MAY_READ, &err)))
		return err;

	if (!inode->i_op || !inode->i_op->readlink)
		return nfserr_io;

	oldfs = get_fs();
	set_fs(get_ds());
	inode->i_count++;
	err = inode->i_op->readlink(inode, buf, *lenp);
	set_fs(oldfs);

	if (err < 0)
		return nfserrno(-err);
	*lenp = err;

	return 0;
}

/*
 * Create a symlink.
 */
int
nfsd_symlink(svc_req *rqstp, svc_fh *fhp, char *fname, int flen,
					  char *path,  int plen)
{
	struct inode *dir;
	int	err;

	if (!flen || !plen)
		return nfserr_noent;

	dir = fh_inode(rqstp, fhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err);
	if (!dir)
		return err;

	if (nfsd_iscovered(dir))
		return nfserr_perm;

	if (!dir->i_op || !dir->i_op->symlink)
		return nfserr_perm;
	dir->i_count++;
	down(&dir->i_sem);
	err = dir->i_op->symlink(dir, fname, flen, path);
	up(&dir->i_sem);

	if (!err && EX_ISSYNC(fhp->fh_export))
		write_inode(dir);

	return err? nfserrno(-err) : 0;
}

/*
 * Create a hardlink
 */
int
nfsd_link(svc_req *rqstp, svc_fh *fhp, char *fname, int len, svc_fh *tofhp)
{
	struct inode	*dir, *dest;
	int		err;

	if (!(dir = fh_inode(rqstp, fhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err)) ||
	    !(dest = fh_inode(rqstp, tofhp, S_IFREG, MAY_NOP, &err)))
		return err;

	if (!len)
		return nfserr_perm;
	if (nfsd_iscovered(dir))
		return nfserr_acces;
	if (dir->i_dev != dest->i_dev)
		return nfserr_acces;	/* FIXME: nxdev for NFSv3 */
	if (IS_IMMUTABLE(dest) || IS_APPEND(dest))
		return nfserr_perm;
	if (!dir->i_op || !dir->i_op->link)
		return nfserr_perm;
	dir->i_count++;
	down(&dir->i_sem);
	err = dir->i_op->link(dest, dir, fname, len);
	up(&dir->i_sem);

	if (!err && EX_ISSYNC(tofhp->fh_export))
		write_inode(dest);

	return err? nfserrno(-err) : 0;
}

/*
 * Rename a file
 */
int
nfsd_rename(svc_req *rqstp, svc_fh *ffhp, char *fname, int flen,
			    svc_fh *tfhp, char *tname, int tlen)
{
	struct inode	*fdir, *tdir;
	int		err;

	if (!(fdir = fh_inode(rqstp, ffhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err))
	 || !(tdir = fh_inode(rqstp, tfhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err)))
		return err;

	if (!flen || (fname[0] == '.' && 
	    (flen == 1 || (flen == 2 && fname[1] == '.'))) ||
	    !tlen || (tname[0] == '.' && 
	    (tlen == 1 || (tlen == 2 && tname[1] == '.'))))
		return nfserr_perm;

	if (nfsd_iscovered(fdir) || nfsd_iscovered(tdir))
		return nfserr_acces;
	if (fdir->i_dev != tdir->i_dev)
		return nfserr_io;
	if (IS_APPEND(fdir))
		return nfserr_perm;
	if (!fdir->i_op || !fdir->i_op->rename)
		return nfserr_perm;

	tdir->i_count++;
	fdir->i_count++;
	down(&tdir->i_sem);
	err = fdir->i_op->rename(fdir, fname, flen, tdir, tname, tlen);
	up(&tdir->i_sem);

	if (!err && EX_ISSYNC(tfhp->fh_export)) {
		write_inode(fdir);
		write_inode(tdir);
	}

	return err? nfserrno(-err) : 0;
}

/*
 * Remove a file
 */
int
nfsd_remove(svc_req *rqstp, svc_fh *fhp, char *fname, int flen)
{
	struct inode	*dir;
	int		err;

	if (!flen)
		return nfserr_perm;
	if (flen < 3 && fname[0] == '.' && (flen == 1 || fname[1] == '.'))
		return nfserr_acces;
	if (!(dir = fh_inode(rqstp, fhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err)))
		return err;
	if (nfsd_iscovered(dir))
		return nfserr_acces;
	if (IS_APPEND(dir))
		return nfserr_perm;
	if (!dir->i_op || !dir->i_op->unlink)
		return nfserr_perm;
	dir->i_count++;
	err = dir->i_op->unlink(dir, fname, flen);

	if (!err && EX_ISSYNC(fhp->fh_export))
		write_inode(dir);

	return err? nfserrno(-err) : 0;
}

/*
 * Create a directory.
 */
int
nfsd_mkdir(svc_req *rqstp, svc_fh *fhp, char *fname, int nlen,
					struct iattr *iap, svc_fh **resfh)
{
	struct inode	*dir;
	svc_fh		*newfhp;
	int		mode = 0, err;

	if (iap->ia_valid & ATTR_MODE)
		mode = iap->ia_mode & S_IALLUGO;

	if (!nlen)
		return nfserr_perm;
	if (!(dir = fh_inode(rqstp, fhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err)))
		return err;
	if (nfsd_iscovered(dir))
		return nfserr_acces;
	if (!dir->i_op || !dir->i_op->mkdir)
		return nfserr_perm;
	dir->i_count++;
	down(&dir->i_sem);
	err = dir->i_op->mkdir(dir, fname, nlen, mode);
	up(&dir->i_sem);

	/* lookup new directory */
	if ((err = nfsd_lookup(rqstp, fhp, fname, nlen, &newfhp)) != 0)
		return err;

	/* Map the client's uid/gids */
	/* nfsd_sattr_umap(fhp->fh_export, &attr); */

	/* To set atime/mtime, we would usually require that the inode
	 * is writable. To allow proper semantics on directories created
	 * with mode 0555, we forgo this check.
	 */
	iap->ia_valid &= (ATTR_ATIME | ATTR_ATIME_SET |
				ATTR_MTIME | ATTR_MTIME_SET);
	if (iap->ia_valid) {
		iap->ia_valid |= ATTR_CTIME;
		iap->ia_ctime = CURRENT_TIME;
		err = nfsd_notify_change(newfhp->fh_inode, iap);
		if (err) {
			fh_put(newfhp);
		} else if (EX_ISSYNC(fhp->fh_export))
			write_inode(newfhp->fh_inode);
	}

	*resfh = newfhp;
	return err? nfserrno(-err) : 0;
}

/*
 * Remove a directory.
 */
int
nfsd_rmdir(svc_req *rqstp, svc_fh *fhp, char *fname, int nlen)
{
	struct inode	*dir;
	int	err;

	if (!nlen)
		return nfserr_perm;
	if (nlen < 3 && fname[0] == '.' && (nlen == 1 || fname[1] == '.'))
		return nfserr_acces;
	if (!(dir = fh_inode(rqstp, fhp, S_IFDIR, MAY_EXEC|MAY_WRITE, &err)))
		return err;
	if (nfsd_iscovered(dir))
		return nfserr_acces;
	if (IS_APPEND(dir))
		return nfserr_perm;
	if (!dir->i_op || !dir->i_op->rmdir)
		return nfserr_perm;
	dir->i_count++;
	err = dir->i_op->rmdir(dir, fname, nlen);

	if (!err && EX_ISSYNC(fhp->fh_export))
		write_inode(dir);

	return err? nfserrno(-err) : 0;
}

/*
 * Read entries from a directory.
 * This is still not optimal, because the client has to place a separate
 * call to read the eof indicator.
 */
int
nfsd_readdir(svc_req *rqstp, svc_fh *fhp, loff_t offset, 
				encode_dent_fn func, svc_buf *resp, int *eof)
{
	struct readdir_cd cd;
	struct file	*filp;
	int		err;

	if (offset > ~(unsigned long)0)
		return 0;

	if ((err = nfsd_open(rqstp, fhp, S_IFDIR, O_RDONLY, 0, &filp)) != 0)
		return nfserrno(-err);

	if (!filp->f_op || !filp->f_op->readdir) {
		nfsd_close (filp);
		return nfserr_notdir;
	}
	filp->f_pos = (off_t) offset;

	/* Read the directory entries. The new 1.3 code allows us to read
	 * all entries in one go.
	 */
	cd.rqstp = rqstp;
	cd.resp = resp;
	cd.offset = NULL;
	cd.eof = eof;

	err = filp->f_op->readdir(filp->f_inode, filp, &cd, (filldir_t) func);
	if (err < 0)
		return nfserrno(-err);

	/* Hewlett & Packard think offset 0 means start reading again
	 * from beginning. Aaaargh! */
	if (cd.offset /* && !eof */)
		*cd.offset = htonl(filp->f_pos);

	nfsd_close(filp);
	return 0;
}

/*
 * Check for a user's access permissions to this inode.
 * This function has too much overhead. The secure port check and the
 * setuser stuff should be moved to the top-level loop, but
 * unfortunately, we currently implement these things as mount options.
 * We may want to have an /etc/nfsclients file one day, though (for the
 * MAC keys anyway).
 */
int
nfsd_permission(svc_req *rqstp, svc_export *exp, struct inode *inode, int acc)
{
	int		err;

	/* Check if the request originated from a secure port. */
	if (!rqstp->rq_secure
	 && !(exp->ex_flags & NFSEXP_INSECURE_PORT)) {
		printk(KERN_WARNING
			"nfsd: request from insecure port (%08lx:%d)!\n",
				ntohl(rqstp->rq_addr.sin_addr.s_addr),
				rqstp->rq_addr.sin_port);
		return nfserr_perm;
	}

	if (acc == MAY_NOP)
		return 0;

	if (acc & (MAY_WRITE | MAY_SATTR)) {
		if (EX_RDONLY(exp) || IS_RDONLY(inode))
			return nfserr_rofs;
		if (IS_IMMUTABLE(inode))
			return nfserr_perm;
	}

	if (!rqstp->rq_userset) {
		struct svc_authops	*aop = rqstp->rq_cred.ah_op;
		if (!aop->setuser || !aop->setuser(rqstp, exp))
			return nfserr_perm;
		rqstp->rq_userset = 1;
	}

	/* The file owner always gets access permission. This is to make
	 * file access work even when the client has done a fchmod(fd, 0).
	 *
	 * FIXME: `cp foo bar' should fail nevertheless when bar is
	 * readonly. For Sun clients this is fairly easy, because
	 * these clients will always start the copy with a creat() call.
	 * Linux doesn't do this, but it still starts with a truncate()
	 * call, for which we could check. Perhaps we need a MAY_TRUNC
	 * condition to check for this.
	 */
	if (inode->i_uid == current->fsuid)
		return 0;

	err = permission(inode, acc);

	/* Allow read access to binaries even when mode 111 */
	if (err == -EPERM && S_ISREG(inode->i_mode) && acc == MAY_READ)
		err = permission(inode, MAY_EXEC);

	return err? nfserrno(-err) : 0;
}

/*
 * Look up the inode for a given FH.
 */
struct inode *
nfsd_iget(dev_t dev, ino_t ino)
{
	struct super_block	*sb;

	if (!(sb = get_super(dev)))
		return NULL;
	return __iget(sb, ino, 0);
}

/*
 * Look up the root inode of the parent fs.
 * We have to go through iget in order to allow for wait_on_inode.
 */
int
nfsd_parentdev(dev_t* devp)
{
	struct super_block	*sb;

	if (!(sb = get_super(*devp)) || !sb->s_covered)
		return 0;
	if (*devp == sb->s_covered->i_dev)
		return 0;
	*devp = sb->s_covered->i_dev;
	return 1;
}

/* Duplicated here from fs/super.c because it's not exported */
static struct super_block *
get_super(dev_t dev)
{
	struct super_block *s;

	if (!dev)
		return NULL;
	s = 0 + super_blocks;
	while (s < NR_SUPER + super_blocks)
		if (s->s_dev == dev) {
			wait_on_super(s);
			if (s->s_dev == dev)
				return s;
			s = 0 + super_blocks;
		} else
			s++;
	return NULL;
}

/*
 * This is a copy from fs/inode.c because it wasn't exported.
 */
int
nfsd_notify_change(struct inode *inode, struct iattr *attr)
{
	int retval;

	if (inode->i_sb && inode->i_sb->s_op &&
	    inode->i_sb->s_op->notify_change)
		return inode->i_sb->s_op->notify_change(inode, attr);

	if ((retval = inode_change_ok(inode, attr)) != 0)
		return retval;

	inode_setattr(inode, attr);
	return 0;
}

