/*
 * nfsctl.c	Syscall interface to knfsd.
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */

#ifdef MODULE
#include <linux/module.h>
#include <linux/modversions.h>
#include <linux/version.h>
#endif

#include <linux/linkage.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/nfs.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>

#include <nfs/nfsctl.h>
#include <nfs/nfsd.h>
#include <nfs/cache.h>
#include <version.h>

#ifdef MODULE
char kernel_version[]=UTS_RELEASE;
#endif

extern long sys_call_table[];

/* int asmlinkage sys_nfs_ctl(int fd); */
static int	nfsctl_svc(nfs_svc *data);
static int	nfsctl_addclient(nfs_client *data);
static int	nfsctl_delclient(nfs_client *data);
static int	nfsctl_export(nfs_export *data);
static int	nfsctl_unexport(nfs_export *data);
static int	nfsctl_getfh(nfsctl_getfh_arg *data, struct knfs_fh *res);
/* static int	nfsctl_ugidupdate(nfs_ugidmap *data); */

unsigned int	nfsd_debug = 0;

static inline int
nfsctl_svc(struct nfs_svc *data)
{
	if (data->svc_nrfds >= NFS_MAXSVCFDS)
		return -EINVAL;
	return nfsd_svc(data->svc_fd, data->svc_nrfds);
}

static inline int
nfsctl_addclient(nfs_client *data)
{
	return exp_addclient(data);
}

static inline int
nfsctl_delclient(nfs_client *data)
{
	return exp_delclient(data);
}

static inline int
nfsctl_export(nfs_export *data)
{
	return exp_export(data);
}

static inline int
nfsctl_unexport(nfs_export *data)
{
	return exp_unexport(data);
}

#ifdef notyet
static inline int
nfsctl_ugidupdate(nfs_ugidmap *data)
{
	return -EINVAL;
}
#endif

#ifdef notyet
static inline int
nfsctl_openfh(nfs_openarg *data)
{
	struct sockaddr_in *sin;
	svc_client	*clp;
	svc_fh		*fhp;
	int		fd, max;

	if (data->op_addr.sa_family != AF_INET)
		return -EINVAL;
	sin = (struct sockaddr_in *)&data->op_addr;

	if ((max = current->rlimit[RLIMIT_NOFILE].rlim_cur) > NR_OPEN)
		max = NR_OPEN;
	for (fd=0; fd < max; fd++)
		if (!current->files->fd[fd])
			break;
	if (fd >= max)
		return -EMFILE;

	exp_readlock();
	if (!(clp = exp_getclient(&sin->sin_addr))) {
		exp_unlock();
		return -EPERM;
	}

	FD_CLR(fd, &current->files->close_on_exec);
	if (!(f = get_empty_filp())) {
		exp_unlock();
		return -ENFILE;
	}

	/* to be filled in */

	exp_unlock();
	return fd;
}
#endif

static inline int
nfsctl_getfh(nfsctl_getfh_arg *data, struct knfs_fh *res)
{
	struct sockaddr_in	*sin;
	svc_client		*clp;
	int			err = 0;

	if (data->gf_addr.sa_family != AF_INET)
		return -EPROTONOSUPPORT;
	if (data->gf_version < 2 || data->gf_version > NFSD_MAXVERS)
		return -EINVAL;
	sin = (struct sockaddr_in *)&data->gf_addr;

	exp_readlock();
	if (!(clp = exp_getclient(&sin->sin_addr)))
		err = -EPERM;
	else
		err = exp_rootfh(clp, data->gf_version, data->gf_path, res);
	printk("nfsd: rootfh ret code %d\n", -err);
	exp_unlock();

	return err;
}

static inline int
nfsctl_authd(struct auth_args *argp, struct auth_args *resp)
{
	/* switch argp and resp: our argument is the reply to an
	 * earlier kernel request, and our result is the new request.
	 */
	return authd_stub(resp, argp);
}

static inline int
nfsctl_debug(unsigned int *newp, unsigned int *oldp)
{
	*oldp = nfsd_debug;
	if (*newp != NFSDBG_NOCHANGE)
		nfsd_debug = *newp;
	return 0;
}

int
asmlinkage sys_nfsctl(int cmd, nfsctl_arg *argp, nfsctl_res *resp)
{
	nfsctl_arg *	arg = NULL;
	nfsctl_res *	res = NULL;
	int		err;

	if (!suser())
		return -EPERM;
	if ((err = verify_area(VERIFY_READ, argp, sizeof(*argp))) != 0)
		return err;
	if (resp && (err = verify_area(VERIFY_WRITE, resp, sizeof(*resp))))
		return err;
	if (!(arg = kmalloc(sizeof(*arg), GFP_USER)) ||
	    (resp && !(res = kmalloc(sizeof(*res), GFP_USER)))) {
		err = -ENOMEM;	/* ??? */
		goto done;
	}
	memcpy_fromfs(arg, argp, sizeof(*argp));
	if (arg->ca_version != NFSCTL_VERSION) {
		printk(KERN_WARNING "nfsd: incompatible version in syscall.\n");
		err = -EINVAL;
		goto done;
	}

	MOD_INC_USE_COUNT;
	switch(cmd) {
	case NFSCTL_SVC:
		err = nfsctl_svc(&arg->ca_svc);
		break;
	case NFSCTL_ADDCLIENT:
		err = nfsctl_addclient(&arg->ca_client);
		break;
	case NFSCTL_DELCLIENT:
		err = nfsctl_delclient(&arg->ca_client);
		break;
	case NFSCTL_EXPORT:
		err = nfsctl_export(&arg->ca_export);
		break;
	case NFSCTL_UNEXPORT:
		err = nfsctl_unexport(&arg->ca_export);
		break;
#ifdef notyet
	case NFSCTL_UGIDUPDATE:
		err = nfsctl_ugidupdate(&arg->ca_umap);
		break;
#endif
	case NFSCTL_GETFH:
		err = nfsctl_getfh(&arg->ca_getfh, &res->cr_getfh);
		break;
	case NFSCTL_AUTHD:
		err = nfsctl_authd(&arg->ca_authd, &res->cr_authd);
		break;
	case NFSCTL_DEBUG:
		err = nfsctl_debug(&arg->ca_debug, &res->cr_debug);
		break;
	default:
		err = -EINVAL;
	}
	MOD_DEC_USE_COUNT;

	if (!err && resp)
		memcpy_tofs(resp, res, sizeof(*resp));

done:
	if (arg)
		kfree(arg);
	if (res)
		kfree(res);

	return err;
}

int
init_module(void)
{
	sys_call_table[__NR_nfsservctl] = (long) sys_nfsctl;
	printk("Installed knfsd %s.\n", VERSION);
	nfsstat_init();
	nfscache_init();
	exp_initialize();
	return 0;
}

void
cleanup_module(void)
{
	if (MOD_IN_USE) {
		printk("nfsd: nfsd busy, remove delayed\n");
		return;
	}
	sys_call_table[__NR_nfsservctl] = 0;
	nfsd_shutdown_sockets();
	exp_shutdown();
	nfscache_shutdown();
	nfsstat_shutdown();
}
