/*
 * nfsd.h		Hodge-podge collection of knfsd-related stuff.
 *			I will sort this out later.
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */

#include <linux/types.h>
#include <linux/unistd.h>
#include <linux/dirent.h>
#include <linux/fs.h>

#include <nfs/nfsfh.h>
#include <nfs/nfsmac.h>
#include <nfs/export.h>
#include <nfs/auth.h>
#include <nfs/stats.h>

#ifndef __NFSD_H__
#define __NFSD_H__

/*
 * knfsd debug flags
 */
#define NFSDBG_SOCK	0x0001
#define NFSDBG_FH	0x0002
#define NFSDBG_EXPORT	0x0004
#define NFSDBG_SVC	0x0008
#define NFSDBG_PROC	0x0010
#define NFSDBG_FILEOP	0x0020
#define NFSDBG_AUTH	0x0040
#define NFSDBG_REPCACHE	0x0080
#define NFSDBG_NOCHANGE	0xFFFF

#ifdef __KERNEL__
/*
 * This is our internal representation of an NFS handle. Having an FH
 * on two doubly linked lists may be overkill memory-wise, but I believe
 * it's worth it in terms of lookup time. An fh cache with 512 entries
 * will account for no more than about 30K on an Intel machine.
 * Alternatively, we could use the index into the static fh store
 * instead of pointers (even 16bit types would suffice for this).
 */
typedef struct svc_fh {
	unsigned short		fh_magic;	/* magic */
	unsigned short		fh_pvers;	/* NFS protocol version */
	struct knfs_fh		fh_handle;	/* FH data */
	svc_export *		fh_export;	/* export pointer */
	struct svc_fh *		fh_next;	/* (ino,dev) hash list */
	struct svc_fh *		fh_prev;
	struct svc_fh *		fh_lru_next;	/* LRU list */
	struct svc_fh *		fh_lru_prev;
	unsigned short		fh_count;	/* refcount */
	unsigned short		fh_used : 1;	/* used recently */
	struct inode *		fh_inode;	/* inode */
} svc_fh;
#define SVCFH_MAGIC		0xbaff

/*
 * NFS socket. This is our encapsulation of TCP/UDP NFS sockets.
 *
 * NOTE: First two items must be prev/next.
 */
typedef struct svc_sock {
	struct svc_sock *	sk_prev;	/* list of ready sockets */
	struct svc_sock *	sk_next;
	struct svc_sock *	sk_list;	/* list of all sockets */
	struct svc_req *	sk_rqstp;	/* XXX: debugging */
	struct socket *		sk_sock;	/* berkeley socket layer */
	struct sock *		sk_sk;		/* INET layer */
	char			sk_inuse;	/* use count */
	char			sk_busy;	/* enqueued for reading */
	char			sk_conn;	/* conn pending */
	char			sk_close;	/* dead or dying */
	int			sk_data;	/* data pending */
	unsigned int		sk_temp : 1;	/* temp socket */
	int			(*sk_recvfrom)(struct svc_req *rqstp);
	int			(*sk_sendto)(struct svc_req *rqstp);

	/* We keep the old state_change and data_ready CB's here */
	void			(*sk_ostate)(struct sock *);
	void			(*sk_odata)(struct sock *, int bytes);

	/* private TCP part */
	unsigned long		sk_reclen;	/* length of record */
	u32 *			sk_rbuf;	/* read buffer */
	int			sk_rlen;	/* read length */
} svc_sock;

/*
 * Buffer to store NFS requests or replies in. Later we want to have a
 * request cache.
 */
typedef struct svc_buf {
	u32 *			base;
	int			buflen;
	u32 *			buf;
	int			len;
} svc_buf;

/* Current limit is 8K read/write buffer */
#define MAXBUF          (8192u+1024u)


/*
 * The context of a single thread, including the request currently being
 * processed.
 * NOTE: First two items must be prev/next.
 */
typedef struct svc_req {
	struct svc_req *	rq_prev;	/* idle list */
	struct svc_req *	rq_next;
	struct svc_sock *	rq_sock;	/* socket */
	struct sockaddr_in	rq_addr;	/* peer address */
	int			rq_addrlen;
	svc_client *		rq_client;	/* NFS peer info */
	svc_auth		rq_cred;	/* auth info */
	svc_buf			rq_args;	/* argument buffer */
	svc_buf			rq_res;		/* result buffer */
	struct wait_queue *	rq_wait;
	struct nfscache_ent *	rq_cacherep;
	u32			rq_xid;		/* transmission id */
	u32			rq_vers;	/* protocol version */
	u32			rq_proc;	/* procedure number */
	unsigned short		rq_userset : 1,	/* auth->setuser OK */
				rq_secure  : 1;	/* secure port */
} svc_req;

/*
 * Procedure table for NFS
 */
struct svc_proc {
	int		(*p_func) (svc_req *, svc_buf *, svc_buf *);
	char		*p_name;
	int		p_count;
	unsigned long	p_time;
	unsigned char	p_cachetype;
};
extern struct svc_proc	nfsd_proc2[18];

/*
 * Version table for NFS
 */
struct svc_version {
	struct svc_proc	*sv_procs;
	int		sv_nproc;
};
extern struct svc_version nfsd_version[NFSD_MAXVERS+1];

/*
 * Special flags for nfsd_permission. These must be different from MAY_READ,
 * MAY_WRITE, and MAY_EXEC.
 */
#define MAY_NOP		0
#define MAY_SATTR	8
#if MAY_SATTR & (MAY_READ | MAY_WRITE | MAY_EXEC)
# error "please use a different value for MAY_SATTR."
#endif

/*
 * Callback function for readdir
 */
/*
typedef int	(*encode_dent_fn)(svc_req*, svc_buf*, struct dirent*, loff_t);
 */
struct readdir_cd {
	svc_req *	rqstp;
	svc_buf *	resp;
	u32 *		offset;
	int *		eof;
};
typedef int	(*encode_dent_fn)(struct readdir_cd *, const char *,
						int, off_t, ino_t);

/*
 * knfsd debugging stuff
 */
extern unsigned int	nfsd_debug;
#define ifdebug(flag)	if (nfsd_debug & NFSDBG_##flag)

/*
 * Function prototypes.
 */
int		nfsd_svc(int fds[], int nrfds);
int		nfsd_recv(svc_req *rqstp);
void		nfsd_drop(svc_req *rqstp);
int		nfsd_send(svc_req *rqstp);
int		nfsd_add_socket(struct socket *);
void		nfsd_shutdown_sockets(void);

int		nfsd_process_req(svc_req *);
u32 *		nfsd_accept_reply(svc_req *, u32);

int		nfsd_lookup(svc_req *, svc_fh *, char *, int, svc_fh **);
int		nfsd_setattr(svc_req *, svc_fh *, struct iattr *);
int		nfsd_open(svc_req *, svc_fh *, int, int, int, struct file **);
void		nfsd_close(struct file *);
int		nfsd_create(svc_req *, svc_fh *, char *, int, int, svc_fh **);
int		nfsd_mknod(struct inode *, char *, int, int, dev_t);
int		nfsd_read(svc_req *, svc_fh *, loff_t, char *, unsigned long *);
int		nfsd_write(svc_req *, svc_fh *, loff_t, char *, unsigned long);
int		nfsd_readlink(svc_req *, svc_fh *, char *, int *);
int		nfsd_symlink(svc_req *, svc_fh *, char *, int, char *, int);
int		nfsd_link(svc_req *, svc_fh *, char *, int, svc_fh *);
int		nfsd_rename(svc_req *, svc_fh *, char *, int,
					svc_fh *, char *, int);
int		nfsd_remove(svc_req *, svc_fh *, char *, int);
int		nfsd_mkdir(svc_req *, svc_fh *, char *, int,
					struct iattr *, svc_fh **);
int		nfsd_rmdir(svc_req *, svc_fh *, char *, int);
int		nfsd_truncate(svc_req *, struct inode *, unsigned long);
int		nfsd_readdir(svc_req *, svc_fh *, loff_t,
					encode_dent_fn, svc_buf *, int *);
int		nfsd_notify_change(struct inode *, struct iattr *);
int		nfsd_permission(svc_req *, svc_export *, struct inode *, int);
void		nfsd_fcache_flush(int);

struct inode *	nfsd_iget(dev_t dev, ino_t ino);
int		nfsd_parentdev(dev_t *devp);

#ifndef DEBUG
struct inode *	fh_inode(svc_req *, svc_fh *, int, int, int *);
void		fh_put(svc_fh *);
#else
struct inode *	_fh_inode(svc_req *, svc_fh *, int, int, int *, char *, int);
void		_fh_put(svc_fh *, char *, int);
#define fh_inode(rqstp, fhp, type, acc, errp)	\
		_fh_inode(rqstp, fhp, type, acc, errp, __FILE__, __LINE__)
#define fh_put(fhp)	\
		_fh_put(fhp, __FILE__, __LINE__)
#endif
svc_fh *	fh_compose(svc_req *, svc_export *, struct inode *);
svc_fh *	fh_compose_root(svc_client *, int, svc_export *);
svc_fh *	fh_validate(svc_req *, struct knfs_fh *);
void		fh_flush(int force);
void		fh_invalidate(svc_export *exp);

void		exp_initialize(void);
void		exp_shutdown(void);
void		exp_readlock(void);
int		exp_writelock(void);
void		exp_unlock(void);
svc_client *	exp_getclient(struct in_addr* addr);
void		exp_putclient(svc_client *clp);
svc_export *	exp_get(svc_client *clp, dev_t dev, ino_t ino);
int		exp_rootfh(svc_client *, int, char *, struct knfs_fh *);
int		nfserrno(int errno);

/* Check if the inode is an NFS mount point. */
static inline int
nfsd_iscovered(struct inode *inode)
{
	return inode->i_mount != 0;
}

/* Release an inode */
#if 0
static inline void
nfsd_iput(struct inode *inode)
{
	iput(inode);
}
#else
#define nfsd_iput(inode)	\
	{ if (!inode->i_count) \
		printk("nfsd: trying to free free inode in %s:%d\n" \
		       "      dev %04x ino %ld, mode %07o\n", \
		       __FILE__, __LINE__, inode->i_dev, \
		       inode->i_ino, inode->i_mode);  \
	  else iput(inode); \
	}
#endif

#include <linux/ext2_fs.h>

/* Get inode generation number */
static inline unsigned long
nfsd_iversion(struct inode *inode)
{
	if (inode->i_sb->s_magic == EXT2_SUPER_MAGIC)
		return inode->u.ext2_i.i_version;
	return 0;
}

#ifndef makedev
#define makedev(maj, min)	(((maj) << 8) | (min))
#endif

/*
 * These variables contain pre-xdr'ed values for faster operation.
 * FIXME: should be replaced by macros for big-endian machines.
 */
extern u32	xdr_zero, xdr_one, xdr_two;

extern u32	rpc_success,
		rpc_prog_unavail,
		rpc_prog_mismatch,
		rpc_proc_unavail,
		rpc_garbage_args,
		rpc_system_err;

extern u32	auth_ok,
		autherr_badcred,
		autherr_rejectedcred,
		autherr_badverf,
		autherr_rejectedverf,
		autherr_tooweak;

extern u32	nfs_ok,
		nfserr_perm,
		nfserr_noent,
		nfserr_io,
		nfserr_nxio,
		nfserr_acces,
		nfserr_exist,
		nfserr_nodev,
		nfserr_notdir,
		nfserr_isdir,
		nfserr_fbig,
		nfserr_nospc,
		nfserr_rofs,
		nfserr_nametoolong,
		nfserr_dquot,
		nfserr_stale;

#define xdr_putlong(resp, val)	\
	{ *((resp)->buf++) = val; (resp)->len++; }
#define xdr_getlong(argp, var)  \
	{ var = *((argp)->buf++); (argp)->len--; }
#define nfsd_putstat(resp, err)	xdr_putlong(resp, err);

#define nfs_luid(rq, uid)	((u32)(uid))
#define nfs_lgid(rq, gid)	((u32)(gid))
#define nfs_ruid(rq, uid)	((u32)(uid))
#define nfs_rgid(rq, gid)	((u32)(gid))

/* if kernel stack page corrupted, commit seppuku */
#if 0
#define CHECKPOINT() \
	{ unsigned long	xxx;						\
	  xxx = *(unsigned long *)current->kernel_stack_page;		\
	  if (xxx != STACK_MAGIC) {					\
		printk("detected kernel stack corruption in %s:%d\n",	\
			__FILE__, __LINE__);				\
		current->signal |= (1 << (SIGKILL - 1));		\
	  }								\
	}
#else
#define CHECKPOINT()	do { } while (0)
#endif

#endif /* __KERNEL__ */

#endif /* __NFSD_H__ */
