/*
 * export.h		Public declarations for NFS exports. The definitions
 *			for the syscall interface are in nfsctl.h
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */

#ifndef NFSEXPORT_H
#define NFSEXPORT_H

#include <linux/types.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/fs.h>

/*
 * Important limits for the exports stuff.
 */
#define NFSCLNT_IDMAX		1024
#define NFSCLNT_ADDRMAX		16
#define NFSCLNT_KEYMAX		32

/*
 * Export flags.
 */
#define NFSEXP_READONLY		0x0001
#define NFSEXP_INSECURE_PORT	0x0002
#define NFSEXP_ROOTSQUASH	0x0004
#define NFSEXP_ALLSQUASH	0x0008
#define NFSEXP_ASYNC		0x0010
#define NFSEXP_UIDMAP		0x0020
#define NFSEXP_SIGNFH		0x0040
#define NFSEXP_KERBEROS		0x0080
#define NFSEXP_SUNSECURE	0x0100
#define NFSEXP_CROSSMNT		0x0200
#define NFSEXP_ALLFLAGS		0x03FF


#ifdef __KERNEL__

/* The following are hashtable sizes and must be powers of 2 */
#define NFSCLNT_EXPMAX		16
#define NFSEXP_FHMAX		32

/* These must match the actual size of uid_t and gid_t */
#define UGID_BITS		16
#define UGID_SHIFT		8
#define UGID_MASK		((1 << UGID_SHIFT) - 1)
#define UGID_NRENTRIES		((1 << (UGID_BITS - UGID_SHIFT)) + 1)
#define UGID_NONE		((unsigned short)-1)

typedef struct svc_uidmap {
	uid_t			*um_ruid[UGID_NRENTRIES];
	uid_t			*um_luid[UGID_NRENTRIES];
	gid_t			*um_rgid[UGID_NRENTRIES];
	gid_t			*um_lgid[UGID_NRENTRIES];
} svc_uidmap;

typedef struct svc_client {
	struct svc_client *	cl_next;
	char			cl_ident[NFSCLNT_IDMAX];
	int			cl_idlen;
	int			cl_naddr;
	struct in_addr		cl_addr[NFSCLNT_ADDRMAX];
	struct svc_uidmap	cl_umap;
	struct svc_export *	cl_export[NFSCLNT_EXPMAX];
	int			cl_fhkeytype;
	int			cl_fhkeylen;
	unsigned char		cl_fhkey[NFSCLNT_KEYMAX];
	int			cl_count;
	int			cl_lock;
	struct wait_queue *	cl_wait;
} svc_client;

typedef struct svc_export {
	struct svc_export *	ex_next;
	char			ex_path[NFS_MAXPATHLEN+1];
	struct svc_export *	ex_parent;
	struct svc_client *	ex_client;
	int			ex_flags;
	struct inode *		ex_inode;
	dev_t			ex_dev;
	ino_t			ex_ino;
	uid_t			ex_anon_uid;
	gid_t			ex_anon_gid;
	struct svc_fh *		ex_fhcache[NFSEXP_FHMAX];
} svc_export;

#define EX_SECURE(exp)		(!((exp)->ex_flags & NFSEXP_INSECURE_PORT))
#define EX_ISSYNC(exp)		(!((exp)->ex_flags & NFSEXP_ASYNC))
#define EX_RDONLY(exp)		((exp)->ex_flags & NFSEXP_READONLY)
#define EX_CROSSMNT(exp)	((exp)->ex_flags & NFSEXP_CROSSMNT)

/*
 * Inline functions for uid/gid mapping.
 */
static inline uid_t	auth_ruid(svc_export *exp, uid_t uid)
{
	uid_t	*uip;

	if (!(uip = exp->ex_client->cl_umap.um_luid[uid >> UGID_SHIFT]))
		return exp->ex_anon_uid;
	return uip[uid & UGID_MASK];
}


static inline gid_t	auth_rgid(svc_export *exp, gid_t gid)
{
	gid_t	*gip;

	if (!(gip = exp->ex_client->cl_umap.um_lgid[gid >> UGID_SHIFT]))
		return exp->ex_anon_gid;
	return gip[gid & UGID_MASK];
}

static inline uid_t	auth_luid(svc_export *exp, uid_t uid)
{
	uid_t	*uip;

	if ((exp->ex_flags & NFSEXP_ALLSQUASH) ||
	    ((exp->ex_flags & NFSEXP_ROOTSQUASH) && !uid))
		return exp->ex_anon_uid;
	if (!(exp->ex_flags & NFSEXP_UIDMAP))
		return uid;

	if ((uid >> UGID_SHIFT) > UGID_NRENTRIES) {
		printk(KERN_WARNING "nfsd: bad luid umap size (uid = %d)!\n",
			uid);
		return exp->ex_anon_uid;
	}
	if (!(uip = exp->ex_client->cl_umap.um_luid[uid >> UGID_SHIFT]))
		return exp->ex_anon_uid;
	return uip[uid & UGID_MASK];
}

static inline gid_t	auth_lgid(svc_export *exp, gid_t gid)
{
	gid_t	*gip;

	if ((exp->ex_flags & NFSEXP_ALLSQUASH) ||
	    ((exp->ex_flags & NFSEXP_ROOTSQUASH) && !gid))
		return exp->ex_anon_gid;
	if (!(exp->ex_flags & NFSEXP_UIDMAP))
		return gid;

	if ((gid >> UGID_SHIFT) > UGID_NRENTRIES) {
		printk(KERN_WARNING "nfsd: bad lgid umap size (gid = %d)!\n",
			gid);
		return exp->ex_anon_gid;
	}
	if (!(gip = exp->ex_client->cl_umap.um_lgid[gid >> UGID_SHIFT]))
		return exp->ex_anon_gid;
	return gip[gid & UGID_MASK];
}
#endif /* __KERNEL__ */

#endif /* NFSEXPORT_H */
