/*
 * nfsfh2.c	NFS filehandle cache and validation.
 *		Much code in this file could be shared between NFSv2 and
 *		NFSv3, but this will have to wait until the great cleanup.
 *
 *		NOTE: Currently, file handles are not shared between
 *		different exports. I *believe* this will accelerate lookups
 *		in the fh cache, but this may be wishful thinking.
 *		It is however required in order to have support different
 *		hash signatures for different clients.
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */

#ifdef MODULE
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/stat.h>
#include "nfsd.h"

#define FHCACHE_SIZE		512
#define NFSEXP_FHMASK		(NFSEXP_FHMAX-1)
#define HASH(dev, ino)		(((dev) ^ (ino)) & NFSEXP_FHMASK)

static svc_fh			fhstore[FHCACHE_SIZE];
static svc_fh *			freelist = NULL;
static svc_fh *			lru_head = NULL;
static svc_fh *			lru_tail = NULL;
static int			initialized = 0;

static int			cookiesize[NFSD_MAXVERS+1] = {
					0,
					0,
					NFS2_FHCOOKIE_SIZE,
#ifdef NFS3_SUPPORTED
					NFS3_FHCOOKIE_SIZE,
#endif
				};

static svc_fh *		fh_lookup(svc_req *rqstp,
						dev_t dev, ino_t ino,
						dev_t xdev, ino_t xino,
						u32 vers, int validate);
static void		fh_init(void);
static svc_fh *		fh_alloc(void);
static void		fh_free(svc_fh **h, svc_fh *fhp, int force);
static void		fh_movehead(svc_fh **h, svc_fh *fhp);
static void		fh_unlink(svc_fh **h, svc_fh *fhp);
static void		fh_sign(svc_req *rqstp, struct knfs_fh *fh);

/*
 * Get the inode given a file handle.
 */
#ifndef DEBUG
struct inode *
fh_inode(svc_req *rqstp, svc_fh *fhp, int type, int access, int *errp)
{
	struct inode *	inode = fhp->fh_inode;

	/* Type check. The correct error return for type mismatches
	 * does not seem to be generally agreed upon. SunOS seems to
	 * use EISDIR if file isn't S_IFREG; a comment in the NFSv3
	 * spec says this is incorrect (implementation notes for the
	 * write call).
	 */
	if (type && (inode->i_mode & S_IFMT) != type) {
		if (type == S_IFDIR)
			*errp = nfserr_notdir;
		else
			*errp = nfserr_isdir;	/* Sun compatible */
		return NULL;
	}

	/* check access permissions */
	*errp = nfsd_permission(rqstp, fhp->fh_export, inode, access);
	if (*errp)
		return NULL;

	return inode;
}
#else
struct inode *
_fh_inode(svc_req *rqstp, svc_fh *fhp, int type, int access, int *errp,
					char *file, int line)
{
	struct inode *	inode = fhp->fh_inode;
	char *		msg;

	/* Get export pointer right away. */
	if (!fhp->fh_export) {
		msg = "NULL export pointer";
		goto fail;
	}

	if (!fhp->fh_count) {
		msg = "fh count 0";
		goto fail;
	}

	if (!inode) {
		msg = "NULL inode pointer";
		goto fail;
	}

	if (!inode->i_count) {
		msg = "inode count 0";
		goto fail;
	}

	/* Type check. The correct error return for type mismatches
	 * does not seem to be generally agreed upon. SunOS seems to
	 * use EISDIR if file isn't S_IFREG; a comment in the NFSv3
	 * spec says this is incorrect (implementation notes for the
	 * write call).
	 */
	if (type && (inode->i_mode & S_IFMT) != type) {
		if (type == S_IFDIR)
			*errp = nfserr_notdir;
		else
			*errp = nfserr_isdir;	/* Sun compatible */
		return NULL;
	}

	/* check access permissions */
	*errp = nfsd_permission(rqstp, fhp->fh_export, inode, access);
	if (*errp)
		return NULL;

	return inode;

fail:
	printk("nfsd %s in %s:%d\n"
	       "     fh: 0x%04x/%ld (ct %d) ino %p (ct %d)\n",
			msg, file, line,
			fhp->fh_handle.fh_dev, fhp->fh_handle.fh_ino,
			fhp->fh_count,
			inode, inode? inode->i_count : -1);
	*errp = nfserr_stale;
	return NULL;
}
#endif

/*
 * Compose file handle for NFS reply.
 */
svc_fh *
fh_compose(svc_req *rqstp, svc_export *exp, struct inode *inode)
{
	u32		version = 0;
	svc_fh		*fhp;

	version = nfsd_iversion(inode);
	fhp = fh_lookup(rqstp, inode->i_dev, inode->i_ino,
				    exp->ex_dev, exp->ex_ino, version, 0);
	if (fhp) {
		if (!fhp->fh_inode)
			fhp->fh_inode = inode;
		else
			nfsd_iput(inode);
	}

	return fhp;
}

/*
 * Compose root file handle on behalf of mountd. Simply fake the svc_req
 * struct (yucky!).
 */
svc_fh *
fh_compose_root(svc_client *clp, int version, svc_export *exp)
{
	struct inode	*inode = exp->ex_inode;
	struct svc_req	rqst;
	svc_fh		*fhp;

	rqst.rq_vers = version;
	rqst.rq_client = clp;

	/* We must frob i_count here because fh_put will try to decrement
	 * it. */
	inode->i_count++;
	if (!(fhp = fh_compose(&rqst, exp, inode)))
		inode->i_count--;
	return fhp;
}

/*
 * Validate file handle given in NFS request.
 */
svc_fh *
fh_validate(svc_req *rqstp, struct knfs_fh *fh)
{
	struct inode	*inode;
	svc_fh		*fhp;
	int		err;

	fhp = fh_lookup(rqstp, fh->fh_dev,  fh->fh_ino,
				fh->fh_xdev, fh->fh_xino,
				fh->fh_version, 1);
	if (!fhp) {
		nfsdstats.fhstale++;
		return NULL;
	}

	if (!(inode = fhp->fh_inode)) {
		if (!(inode = nfsd_iget(fh->fh_dev, fh->fh_ino)))
			return NULL;
		if (!inode->i_nlink || fh->fh_version != nfsd_iversion(inode)) {
			/* unlinked in the meanwhile */
			nfsd_iput(inode);
			nfsdstats.fhstale++;
			return NULL;
		}
		wait_on_inode(inode);
		if (!fhp->fh_inode)
			fhp->fh_inode = inode;
		else
			nfsd_iput(inode);
	}

	err = memcmp(fhp->fh_handle.fh_cookie, fh->fh_cookie,
			cookiesize[rqstp->rq_vers]);
	if (err) {
		printk(KERN_WARNING
			"nfsd: invalid file handle from %08lx:%d.\n",
				ntohl(rqstp->rq_addr.sin_addr.s_addr),
				ntohs(rqstp->rq_addr.sin_port));
		nfsdstats.fhbadcksum++;
		fh_put(fhp);
		return NULL;
	}

	return fhp;
}

#ifndef DEBUG
void
fh_put(svc_fh *fhp)
{
	if (!fhp || !fhp->fh_count)
		return;
	fhp->fh_count--;
}
#else
void
_fh_put(svc_fh *fhp, char *file, int line)
{
	struct inode *	inode;
	char *		msg;

	if (!fhp) {
		printk(KERN_WARNING "fh_put: fhp == NULL in %s:%d\n",
				file, line);
		return;
	}
	inode = fhp->fh_inode;
	if (!fhp->fh_count) {
		msg = "fh_count == 0";
		goto fail;
	}
	if (inode != NULL && inode->i_count == 0) {
		fhp->fh_inode = NULL;
		fhp->fh_count--;
		msg = "inode count == 0";
		goto fail;
	}
	fhp->fh_count--;
	return;

fail:
	printk("fh_put: %s in %s:%d\n"
	       "     fh: 0x%04x/%ld (ct %d) ino %p (ct %d)\n",
			msg, file, line,
			fhp->fh_handle.fh_dev, fhp->fh_handle.fh_ino,
			fhp->fh_count,
			inode, inode? inode->i_count : -1);
}
#endif

static svc_fh *
fh_lookup(svc_req *rqstp, dev_t dev, ino_t ino, dev_t xdev, ino_t xino,
					u32 vers, int validate)
{
	struct knfs_fh	*fh = NULL;	/* compiler food */
	struct inode	*inode;
	svc_export	*exp;
	svc_fh		*fhp, **h;

	if (!initialized)
		fh_init();

	if (!(exp = exp_get(rqstp->rq_client, xdev, xino)))
		return NULL;	/* export entry revoked */

	h = exp->ex_fhcache + HASH(dev, ino);
	for (fhp = *h; fhp != NULL; fhp = fhp->fh_next) {
		if (fhp->fh_magic != SVCFH_MAGIC) {
			printk(KERN_WARNING "bad magic (%s:%d) on q %ld\n",
				__FILE__, __LINE__, HASH(dev, ino));
			continue;
		}
		fh = &(fhp->fh_handle);
		if (fh->fh_dev == dev && fh->fh_ino == ino)
			break;
	}

	if (fhp != NULL) {
		if (fh->fh_xdev != xdev || fh->fh_xino != xino)
			return NULL;	/* export entry revoked */

		if ((inode = fhp->fh_inode) != NULL && inode->i_count == 0)
			printk("fh_lookup: fh %p icount 0\n", fhp);

		/* We have a slight problem here in that an inode's version
		 * may just have changed, but the cache information is stale.
		 * We can't just delete the cache entry, because some other
		 * nfsd may be operating on it. We may however safely update
		 * the cached version number if the client-provided number
		 * matches that of the real inode, or if the inode pointer
		 * is NULL (if that should be wrong, we will detect it as
		 * soon as the caller tries to look up the inode).
		 */
		if (fh->fh_version == vers ||
		    (inode = fhp->fh_inode) == NULL ||
		    nfsd_iversion(inode) == vers) {
			fh_movehead(h, fhp);
			fhp->fh_handle.fh_version = vers;
			fhp->fh_count++;
			fhp->fh_used = 1;
			nfsdstats.fhhits++;
			return fhp;
		}

		/* version number is definitely stale */
		return NULL;
	}

	if (!(fhp = fh_alloc())) {
		printk(KERN_WARNING "nfsd: fh cache overflow!\n");
		return NULL;
	}

	fhp->fh_export = exp;

	fhp->fh_used = 1;
	fhp->fh_count = 1;
	fhp->fh_prev = NULL;
	fhp->fh_next = NULL;
	fhp->fh_lru_prev = NULL;
	fhp->fh_lru_next = NULL;
	fhp->fh_inode = NULL;

	fh = &(fhp->fh_handle);
	fh->fh_dev = dev;
	fh->fh_ino = ino;
	fh->fh_version = vers;
	fh->fh_xdev = xdev;
	fh->fh_xino = xino;
	fh_sign(rqstp, fh);

	fh_movehead(h, fhp);

	nfsdstats.fhmisses++;
	return fhp;
}

/*
 * Sign the NFS file handle. The message authentication code is made
 * up of the fh base and a secret key. This protects us from fh
 * guessing attacks attacks (somewhat). We don't use any the client's
 * address in the hash, because the client may have several addresses,
 * and modern named's reorder them at will.
 */
static void
fh_sign(svc_req *rqstp, struct knfs_fh *fh)
{
	unsigned char 	*cookie = fh->fh_cookie;
	svc_client	*clp = rqstp->rq_client;
	int		size = cookiesize[rqstp->rq_vers];

	if (size > 16)
		memset(cookie + 16, 0, size - 16);

	switch (clp->cl_fhkeytype) {
	default:
		printk(KERN_WARNING "nfsd: unknown signature type!\n");
		printk(KERN_WARNING "nfsd: Using NULL signatures.\n");
		clp->cl_fhkeytype = NFSSIG_NULL;
		/* fallthru */
	case NFSSIG_NULL:
		memcpy(cookie, "please spoof me!", size);
		break;
	case NFSSIG_MD5:
		{
			struct MD5Context	ctxt;

			MD5Init(&ctxt);
			MD5Update(&ctxt, clp->cl_fhkey, clp->cl_fhkeylen);
			MD5Update(&ctxt, (u8 *)&(fh->fh_base),
						sizeof(fh->fh_base));
			MD5Update(&ctxt, clp->cl_ident, clp->cl_idlen);
			if (size >= 16) {
				MD5Final(cookie, &ctxt);
			} else {
				u8	temp[16];

				MD5Final(temp, &ctxt);
				memcpy(cookie, temp, size);
			}
		}
		nfsdstats.fhcksum++;
		break;
	case NFSSIG_SHA:
		{
			struct SHA_INFO		ctxt;

			sha_init(&ctxt);
			sha_update(&ctxt, clp->cl_fhkey, clp->cl_fhkeylen);
			sha_update(&ctxt, (u8 *)&(fh->fh_base),
						sizeof(fh->fh_base));
			sha_update(&ctxt, clp->cl_ident, clp->cl_idlen);
			sha_final(&ctxt);
			memcpy(cookie, ctxt.digest, size);
		}
		nfsdstats.fhcksum++;
		break;
	}
}

static inline void
fh_unlink(svc_fh **h, svc_fh *fhp)
{
	svc_fh *	prev = fhp->fh_prev;
	svc_fh *	next = fhp->fh_next;

	if (prev && prev->fh_magic != SVCFH_MAGIC) {
		printk(KERN_WARNING "unlink: bad magic in prev = %p\n", prev);
		prev = NULL;
	}
	if (next && next->fh_magic != SVCFH_MAGIC) {
		printk(KERN_WARNING "unlink: bad magic in next = %p\n", next);
		next = NULL;
	}
	if (prev)
		prev->fh_next = next;
	else
		*h = next;
	if (next)
		next->fh_prev = prev;
}

static inline void
fh_lru_unlink(svc_fh *fhp)
{
	svc_fh *	prev = fhp->fh_lru_prev;
	svc_fh *	next = fhp->fh_lru_next;

	if (prev && prev->fh_magic != SVCFH_MAGIC) {
		printk(KERN_WARNING "lru_unlink: bad magic in prev = %p\n", prev);
		prev = NULL;
	}
	if (next && next->fh_magic != SVCFH_MAGIC) {
		printk(KERN_WARNING "lru_unlink: bad magic in next = %p\n", next);
		next = NULL;
	}
	if (prev)
		prev->fh_lru_next = next;
	else if (lru_head == fhp)
		lru_head = next;
	if (next)
		next->fh_lru_prev = prev;
	else if (lru_tail == fhp)
		lru_tail = prev;
}

/*
 * Move fh to head of export's fh cache and global LRU list.
 */
static void
fh_movehead(svc_fh **h, svc_fh *fhp)
{
	svc_fh *	tmp;

	if (*h != fhp) {
		fh_unlink(h, fhp);

		fhp->fh_prev = NULL;
		if ((fhp->fh_next = tmp = *h))
			tmp->fh_prev = fhp;
		*h = fhp;
	}

	fh_lru_unlink(fhp);

	fhp->fh_lru_prev = NULL;
	if ((fhp->fh_lru_next = tmp = lru_head) != NULL)
		tmp->fh_lru_prev = fhp;
	lru_head = fhp;
	if (!lru_tail)
		lru_tail = fhp;
}

static void
fh_free(svc_fh **h, svc_fh *fhp, int force)
{
	struct inode *	inode;

	if (fhp->fh_magic != SVCFH_MAGIC) {
		printk(KERN_WARNING "nfsd: bad magic in fh_free!\n");
		return;
	}
	/* FIXME: Maybe we should wait until fh_count drops to zero */
	if (fhp->fh_count) {
		if (!force) {	/* This avoids a yet-unfound bug. */
			printk(KERN_WARNING
				"nfsd: freed fh with count != 0 - not removed\n");
			return;
		}
		printk(KERN_WARNING
			"nfsd: freed fh with count != 0 - removal forced\n");
	}
	if (!fhp->fh_export) {
		printk(KERN_WARNING "nfsd: export pointer NULL in fh_free!\n");
		return;
	}
	if (!h)
		h = fhp->fh_export->ex_fhcache +
			HASH(fhp->fh_handle.fh_dev, fhp->fh_handle.fh_ino);

	fh_lru_unlink(fhp);
	fh_unlink(h, fhp);

	if ((inode = fhp->fh_inode) != NULL) {
		fhp->fh_inode = NULL;
		nfsd_iput(inode);
	}

	fhp->fh_next = freelist;
	freelist = fhp;
}

/*
 * Find a free cache entry. If there are no entries on the free list
 * anymore, we reclaim the least recently used entry.
 */
static svc_fh *
fh_alloc(void)
{
	svc_export	*exp;
	svc_fh		*fhp, **h;
	struct nfs2_fh	*fh;
	struct inode	*inode;

	if ((fhp = freelist) != NULL) {
		freelist = fhp->fh_next;
		if (fhp->fh_magic != SVCFH_MAGIC) {
			printk(KERN_WARNING "nfsd: bad magic in fh_alloc (freelist=%p)!\n", freelist);
			return NULL;
		}
		return fhp;
	}

	for (fhp = lru_tail; fhp && fhp->fh_count; fhp = fhp->fh_lru_prev)
		;

	if (!fhp) {
		printk(KERN_WARNING "nfsd: ran out of fh\'s!\n");
		return NULL;
	}

	if (fhp->fh_magic != SVCFH_MAGIC) {
		printk(KERN_WARNING "nfsd: bad magic in fh_alloc (LRU=%p)!\n", lru_tail);
		return NULL;
	}

	if (!(exp = fhp->fh_export)) {
		printk(KERN_WARNING "nfsd: fh_export = NULL in fh_alloc!\n");
		return NULL;
	}

	fh = (struct nfs2_fh *) &(fhp->fh_handle);
	h = exp->ex_fhcache + HASH(fh->fh_dev, fh->fh_ino);
	fh_lru_unlink(fhp);
	fh_unlink(h, fhp);

	if ((inode = fhp->fh_inode) != NULL) {
		fhp->fh_inode = NULL;
		nfsd_iput(inode);
	}

	return fhp;
}

/*
 * Release all inodes not recently accessed.
 */
void
fh_flush(int force)
{
	struct inode	*inode;
	svc_fh		*fhp;
	int		i, cnt = 0;

	if (!initialized)
		return;

	for (i = 0, fhp = fhstore; i < FHCACHE_SIZE; i++, fhp++) {
		if (!fhp->fh_count && (force || !fhp->fh_used) &&
		    (inode = fhp->fh_inode) != NULL) {
			inode = fhp->fh_inode;
			fhp->fh_inode = NULL;
			nfsd_iput(inode);
			cnt++;
		}
		fhp->fh_used = 0;
	}
}

/*
 * Invalidate all fh's for a given export. The export entry has already
 * been removed from the client's list of exported fs's. Note that because
 * of this, it should be impossible for a file handle in the export's cache
 * to have a count != 0, hence we may safely call fh_free.
 */
void
fh_invalidate(svc_export *exp)
{
	svc_fh		**h;
	int		i;

	if (!initialized)
		return;

	for (i = 0; i < NFSEXP_FHMAX; i++) {
		h = exp->ex_fhcache + i;

		while (*h)
			fh_free(h, *h, 1);
	}
}

static void
fh_init(void)
{
	int		i;

	for (i = 0; i < FHCACHE_SIZE; i++) {
		fhstore[i].fh_next = fhstore + i + 1;
		fhstore[i].fh_magic = SVCFH_MAGIC;
		fhstore[i].fh_inode = NULL;
	}
	fhstore[FHCACHE_SIZE-1].fh_next = NULL;
	freelist = fhstore;

	initialized = 1;
}
