/*
 * nfsproc2.c	Process version 2 NFS requests.
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */

#ifdef MODULE
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/linkage.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/in.h>
/* #include <linux/nfs.h> */
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>

#include "nfsd.h"
#include <nfs/cache.h>

static int	nfsd_proc_null (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_getattr (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_setattr (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_lookup (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_readlink (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_read (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_write (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_create (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_remove (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_rename (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_link (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_symlink (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_mkdir (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_rmdir (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_readdir (svc_req *, svc_buf *, svc_buf *);
static int	nfsd_proc_statfs (svc_req *, svc_buf *, svc_buf *);

static int	nfsd_encode_fh (svc_req *rqstp, svc_buf *resp, svc_fh *fhp);
static int	nfsd_encode_entry (struct readdir_cd *cd, const char *name,
					int namlen, off_t offset, ino_t ino);
static int	nfsd_encode_diropokres(svc_req *rqstp, svc_buf *resp,
					svc_fh *fhp);
static int	nfsd_encode_fattr (svc_req * rqstp, svc_buf * resp,
					struct inode *inode);
static int	nfsd_decode_fh(svc_req * rqstp, svc_buf * argp, svc_fh ** fhp);
static int	nfsd_decode_filename(svc_req * rqstp, svc_buf * argp,
					char **namep, int *flen);
static int	nfsd_decode_diropargs(svc_req *rqstp, svc_buf *argp,
					svc_fh **fhpp, char **namep,
					int *nlen);
static int	nfsd_decode_sattr(svc_req *rqstp, svc_buf *argp,
					struct iattr *iap);

#define sleep(msec)		\
	{	printk(KERN_NOTICE "nfsd: sleeping %d msecs\n", msec); \
		current->state = TASK_INTERRUPTIBLE;	\
		current->timeout = jiffies + msec / 10;	\
		schedule();	\
	}
#define DECODE_FH(fhp)						\
	if (!nfsd_decode_fh(rqstp, argp, &fhp))			\
		return rpc_garbage_args;			\
	if (!fhp)						\
		RETURN(nfserr_stale);
#define DECODE_DIROPARGS(fhp, fname, len)			\
	DECODE_FH(fhp);						\
	if (!nfsd_decode_filename(rqstp, argp, &fname, &len)) {\
		fh_put(fhp); 					\
		return rpc_garbage_args;			\
	}
#define DECODE_DIROPARGS2(fhp, fname, len, ofhp)		\
	DECODE_FH(fhp);						\
	if (!nfsd_decode_filename(rqstp, argp, &fname, &len)) {	\
		fh_put(fhp); 					\
		fh_put(ofhp);					\
		return rpc_garbage_args;			\
	}
#define DECODE_CREATEARGS(fhp, fname, len, attr)		\
	DECODE_DIROPARGS(fhp, fname, len);			\
	if (!nfsd_decode_sattr(rqstp, argp, &attr)) {		\
		fh_put(fhp);					\
		return rpc_garbage_args;			\
	}
#define RETURN(st)	{ nfsd_putstat(resp, st); return rpc_success; }

static u32	nfs_ftypes[] = {
	NFNON,  NFFIFO, NFCHR, NFBAD,
	NFDIR,  NFBAD,  NFBLK, NFBAD,
	NFREG,  NFBAD,  NFLNK, NFBAD,
	NFSOCK, NFBAD,  NFLNK, NFBAD,
};

/*
 * Decode a file handle. Most of the work is done by fh_validate.
 */
static inline int
nfsd_decode_fh(svc_req *rqstp, svc_buf *argp, svc_fh **fhpp)
{
	if (argp->len < NFS2_FHSIZEQ)
		return 0;

	*fhpp = fh_validate(rqstp, (struct knfs_fh *) argp->buf);

	argp->buf += NFS2_FHSIZEQ;
	argp->len -= NFS2_FHSIZEQ;

	return 1;
}

/*
 * Decode a filename argument. We implement a little hack here to avoid a
 * string copy in 3 out of 4 cases. OTOH, we have to do a strlen anyway,
 * so the gain may not be that great.
 * If we were serious, we could even check for NUL bytes and slashes during
 * the long-word copy (cf. (x - 0x01010101) & ((~x) & 0x80808080)).
 */
static inline int
nfsd_decode_filename(svc_req *rqstp, svc_buf *argp, char **namep, int *flen)
{
	u32	*buf = argp->buf;
	char	*nbuf;
	int	len = argp->len;
	int	fl, flq, i;

	if (len-- < 1)
		return 0;

	fl  = ntohl (*buf++);
	flq = (fl + 3) >> 2;
	if (flq > len || !flq)
		return 0;

	if ((fl & 3) != 0) {
		nbuf = (char *) buf;
	} else if (fl <= NFS2_MAXNAMLEN) {
		/* simply shift string forward in buffer */
		u32	*t1 = buf - 1, *t2 = buf;

		nbuf = (char *) (buf - 1);
		for (i = 0; i < flq; i++) 
			*t1++ = *t2++;
	} else {
		return 0;
	}

	/* We don't trust the string given by the sender. */
	for (i = 0; i < fl ; i++) {
		if (!nbuf[i] || nbuf[i] == '/') {
			printk(KERN_NOTICE 
				"nfsd: client supplied bad filename.");
			return 0;
		}
	}

	(*namep = nbuf)[i] = 0;
	*flen = i;

	argp->len = len - flq;
	argp->buf = buf + flq;

	return 1;
}

/*
 * Decode dir operation args (fh + name).
 */
static inline int
nfsd_decode_diropargs(svc_req *rqstp, svc_buf *argp, svc_fh **fhpp,
				char **namep, int *nlen)
{
	if (!nfsd_decode_fh(rqstp, argp, fhpp))
		return 0;
	if (!*fhpp)
		return 1;

	/* Get file name */
	if (!nfsd_decode_filename(rqstp, argp, namep, nlen)) {
		fh_put(*fhpp);
		return 0;
	}
	return 1;
}

/*
 * Decode file attributes to be set.
 */
static inline int
nfsd_decode_sattr(svc_req *rqstp, svc_buf *argp, struct iattr *iap)
{
	u32	*buf;
	u32	tmp, tmp1;

	if (argp->len < 8)
		return 0;
	buf = argp->buf;

	iap->ia_valid = 0;

	/* Sun client bug compatibility check: some sun clients seem to
	 * put 0xffff in the mode field when they mean 0xffffffff.
	 * Quoting the 4.4BSD nfs server code: Nah nah nah nah na nah.
	 */
	if ((tmp = ntohl(*buf++)) != (u32)-1 && tmp != 0xffff) {
		iap->ia_valid |= ATTR_MODE;
		iap->ia_mode = tmp;
	}
	if ((tmp = ntohl(*buf++)) != (u32)-1) {
		iap->ia_valid |= ATTR_UID;
		iap->ia_uid = tmp;
	}
	if ((tmp = ntohl(*buf++)) != (u32)-1) {
		iap->ia_valid |= ATTR_GID;
		iap->ia_gid = tmp;
	}
	if ((tmp = ntohl(*buf++)) != (u32)-1) {
		iap->ia_valid |= ATTR_SIZE;
		iap->ia_size = tmp;
	}
	tmp  = ntohl(*buf++); tmp1 = ntohl(*buf++);
	if (tmp != (u32)-1 && tmp1 != (u32)-1) {
		iap->ia_valid |= ATTR_ATIME | ATTR_ATIME_SET;
		iap->ia_atime = tmp;
	}
	tmp  = ntohl(*buf++); tmp1 = ntohl(*buf++);
	if (tmp != (u32)-1 && tmp1 != (u32)-1) {
		iap->ia_valid |= ATTR_MTIME | ATTR_MTIME_SET;
		iap->ia_mtime = tmp;
	}

	argp->len += buf - argp->buf;
	argp->buf = buf;
	return 1;
}

static inline int
nfsd_encode_fh(svc_req * rqstp, svc_buf * resp, svc_fh * fhp)
{
	memset(resp->buf, 0, NFS2_FHSIZE);
	memcpy(resp->buf, &(fhp->fh_handle), NFS2_FHSIZE);
	resp->buf += NFS2_FHSIZEQ;
	resp->len += NFS2_FHSIZEQ;

	return 1;
}

static inline int
nfsd_encode_fattr(svc_req * rqstp, svc_buf * resp, struct inode *inode)
{
	u32 *bufp = resp->buf;

	if (!inode)
		return 0;
	*bufp++ = htonl(nfs_ftypes[(inode->i_mode & S_IFMT) >> 12]);
	*bufp++ = htonl((u32) inode->i_mode);
	*bufp++ = htonl((u32) inode->i_nlink);
	*bufp++ = htonl((u32) nfs_ruid(rqstp, inode->i_uid));
	*bufp++ = htonl((u32) nfs_rgid(rqstp, inode->i_gid));
	if (S_ISLNK(inode->i_mode)) {
		*bufp++ = htonl(NFS2_MAXPATHLEN);
	} else {
		*bufp++ = htonl((u32) inode->i_size);
	}
	*bufp++ = htonl((u32) inode->i_blksize);
	*bufp++ = htonl((u32) inode->i_rdev);
	*bufp++ = htonl((u32) inode->i_blocks);
	*bufp++ = htonl((u32) inode->i_dev);
	*bufp++ = htonl((u32) inode->i_ino);
	*bufp++ = htonl((u32) inode->i_atime);
	*bufp++ = 0;
	*bufp++ = htonl((u32) inode->i_mtime);
	*bufp++ = 0;
	*bufp++ = htonl((u32) inode->i_ctime);
	*bufp++ = 0;

	resp->buf = bufp;
	resp->len += 17;

	return 1;
}

/*
 * Encode OK result to directory operation.
 */
static inline int
nfsd_encode_diropokres(svc_req *rqstp, svc_buf *resp, svc_fh *fhp)
{
	struct inode *	inode;

	if (!(inode = fhp->fh_inode))
		return 0;
	nfsd_putstat(resp, nfs_ok);
	return nfsd_encode_fh(rqstp, resp, fhp)
	    && nfsd_encode_fattr(rqstp, resp, inode);
}

static inline int
nfsd_encode_entry(struct readdir_cd *cd, const char *name, int namlen,
					off_t offset, ino_t ino)
{
	svc_buf	*resp = cd->resp;
	u32	*buf = resp->buf;
	int	*eof = cd->eof;
	int	slen;

	if (offset > ~((u32) 0))
		offset = ~((u32) 0);
	if (cd->offset)
		*cd->offset = htonl(offset);
	if (namlen > NFS2_MAXNAMLEN) {
		printk (KERN_WARNING "nfsd: filename too long (%d).\n",
							namlen);
		cd->offset = NULL;
		return 0;		/* skip entry */
	}

	slen = (namlen + 3) >> 2;
	if (resp->len + slen + 4 > resp->buflen) {
		*eof = 0;
		return -EINVAL;
	}
	*buf++ = xdr_one;		/* mark entry present */
	*buf++ = htonl((u32) ino);	/* file id */
	*buf++ = htonl((u32) namlen);	/* name length & name */
	memcpy(buf, name, namlen);
	buf += slen;
	cd->offset = buf;		/* remember pointer */
	*buf++ = 0;			/* offset of next entry */

	resp->len += slen + 4;
	resp->buf = buf;

	return 0;
}

/*
 * NFSv2 Server procedures.
 * Only the results of non-idempotent operations are cached.
 */
struct svc_proc		nfsd_proc2[18] = {
	{ nfsd_proc_null,	"null",		0, 0, RC_NOCACHE,  },
	{ nfsd_proc_getattr,	"getattr",	0, 0, RC_NOCACHE,  },
	{ nfsd_proc_setattr,	"setattr",	0, 0, RC_REPLBUFF, },
	{ NULL, 		"root",		0, 0, RC_NOCACHE,  },
	{ nfsd_proc_lookup,	"lookup",	0, 0, RC_NOCACHE,  },
	{ nfsd_proc_readlink,	"readlink",	0, 0, RC_NOCACHE,  },
	{ nfsd_proc_read,	"read",		0, 0, RC_NOCACHE,  },
	{ NULL, 		"writecache",	0, 0, RC_NOCACHE,  },
	{ nfsd_proc_write,	"write",	0, 0, RC_REPLBUFF, },
	{ nfsd_proc_create,	"create",	0, 0, RC_REPLBUFF, },
	{ nfsd_proc_remove,	"remove",	0, 0, RC_REPLSTAT, },
	{ nfsd_proc_rename,	"rename",	0, 0, RC_REPLSTAT, },
	{ nfsd_proc_link,	"link",		0, 0, RC_REPLSTAT, },
	{ nfsd_proc_symlink,	"symlink",	0, 0, RC_REPLSTAT, },
	{ nfsd_proc_mkdir,	"mkdir",	0, 0, RC_REPLBUFF, },
	{ nfsd_proc_rmdir,	"rmdir",	0, 0, RC_REPLSTAT, },
	{ nfsd_proc_readdir,	"readdir",	0, 0, RC_NOCACHE,  },
	{ nfsd_proc_statfs,	"statfs",	0, 0, RC_NOCACHE,  },
};

static int
nfsd_proc_null (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	return rpc_success;
}

static int
nfsd_proc_getattr(svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	struct inode	*inode;
	svc_fh		*fhp;
	int		nfserr;

	DECODE_FH(fhp);

	if (!(inode = fh_inode(rqstp, fhp, 0, MAY_NOP, &nfserr))) {
		fh_put(fhp);
		RETURN(nfserr);
	}

	nfsd_putstat(resp, nfs_ok);
	if (!nfsd_encode_fattr(rqstp, resp, fhp->fh_inode)) {
		fh_put(fhp);
		return rpc_system_err;
	}

	fh_put(fhp);
	return rpc_success;
}

static int
nfsd_proc_setattr(svc_req *rqstp, svc_buf *argp, svc_buf *resp)
{
	struct iattr	attr;
	svc_fh		*fhp;
	int		err;

	DECODE_FH(fhp);

	if (!nfsd_decode_sattr(rqstp, argp, &attr)) {
		fh_put(fhp);
		return rpc_garbage_args;
	}

	/* Change the attributes. */
	if ((err = nfsd_setattr(rqstp, fhp, &attr)) != 0) {
		fh_put(fhp);
		RETURN(err);
	}

	/* Encode new attributes in reply. */
	nfsd_putstat(resp, nfs_ok);
	if (!nfsd_encode_fattr(rqstp, resp, fhp->fh_inode))
		err = rpc_system_err;
	else 
		err = rpc_success;

	fh_put(fhp);
	return err;
}

static int
nfsd_proc_lookup(svc_req *rqstp, svc_buf *argp, svc_buf *resp)
{
	char		*fname;
	svc_fh		*fhp, *newfhp;
	int		nlen, err;

	DECODE_DIROPARGS(fhp, fname, nlen);

	/* Look up the file...  */
	if ((err = nfsd_lookup(rqstp, fhp, fname, nlen, &newfhp)) != 0) {
		fh_put(fhp);
		RETURN(err);
	}

	/* ... and encode the lot. */
	if (!nfsd_encode_diropokres(rqstp, resp, newfhp))
		err = rpc_system_err;
	else
		err = rpc_success;
	fh_put(newfhp);
	fh_put(fhp);
	return err;
}

/*
 * Read a symlink.
 */
static int
nfsd_proc_readlink (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	svc_fh		*fhp;
	int		left;
	int		err, plen;

	DECODE_FH(fhp);

	left = resp->buflen - resp->len;
	if (left < 2 + ((NFS2_MAXPATHLEN + 3) >> 2))
		return rpc_system_err;

	/* Read the symlink. nfsd_readlink puts the inode. */
	plen = NFS2_MAXPATHLEN;
	err = nfsd_readlink(rqstp, fhp, (char *) (resp->buf + 2), &plen);
	fh_put(fhp);

	if (err)
		RETURN(err);

	nfsd_putstat(resp, nfs_ok);
	*(resp->buf++) = htonl(plen);
	resp->len += 1 + ((plen + 3) >> 2);

	return rpc_success;
}

/*
 * Read a portion of a file.
 */
static int
nfsd_proc_read(svc_req *rqstp, svc_buf *argp, svc_buf *resp)
{
	svc_fh		*fhp;
	u32		count, totalcount;
	unsigned long	rcount;
	int		dspace;
	loff_t		offset;
	int		err;

	DECODE_FH(fhp);

	if (argp->len < 3) {
		fh_put(fhp);
		return rpc_garbage_args;
	}

	offset = ntohl(*(argp->buf++));
	count = ntohl(*(argp->buf++));
	totalcount = ntohl(*(argp->buf++));	/* Does anybody use this? */

	/* printk("read %ld bytes at %ld\n", count, offset); */
	/* Compute room for payload. 19 is 1 word for status, 17 words for
	 * fattr, and 1 word for the byte count returned by read().
	 */
	dspace = resp->buflen - resp->len - 19;
	if (dspace < 0 || (dspace << 2) < count) {
		printk("Oops1 (dspace = %d buflen = %d len = %d)\n",
				dspace, resp->buflen, resp->len);
		fh_put(fhp);
		return rpc_system_err;
	}

	rcount = count;
	err = nfsd_read(rqstp, fhp, offset, (char *) (resp->buf+19), &rcount);
	if (err) {
		printk("Oops2\n");
		fh_put(fhp);
		RETURN(err);
	}

	nfsd_putstat(resp, nfs_ok);
	nfsd_encode_fattr(rqstp, resp, fhp->fh_inode);
	*(resp->buf++) = htonl(rcount);
	resp->len += ((rcount + 3) >> 2) + 1;
	/* printk("read got %ld bytes, resp->len %ld\n", rcount, resp->len); */

	fh_put(fhp);
	return rpc_success;
}

static int
nfsd_proc_write (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	svc_fh		*fhp;
	u32		beginoffset, offset,
			totalcount, count;
	int		err;

	DECODE_FH(fhp);

	if ((argp->len -= 4) < 0) {
		fh_put(fhp);
		return rpc_garbage_args;
	}
	beginoffset = ntohl(*(argp->buf++));
	offset = ntohl(*(argp->buf++));
	totalcount = ntohl(*(argp->buf++));
	count = ntohl(*(argp->buf++));

	if (((count + 3) >> 2) > argp->len) {
		fh_put(fhp);
		return rpc_garbage_args;
	}

	/* Make sure there's enough room for nfs_stat and fattr in the reply
	 * buffer.
	 */
	if (resp->buflen - resp->len < 18) {
		fh_put(fhp);
		return rpc_system_err;
	}

	err = nfsd_write(rqstp, fhp, offset, (char *) argp->buf, count);
	if (err) {
		fh_put(fhp);
		RETURN(err);
	}

	nfsd_putstat(resp, nfs_ok);
	nfsd_encode_fattr(rqstp, resp, fhp->fh_inode);

	fh_put(fhp);
	return rpc_success;
}

static int
nfsd_proc_create(svc_req *rqstp, svc_buf *argp, svc_buf *resp)
{
	struct inode	*dirp, *inode = NULL;
	struct iattr	attr;
	svc_fh		*fhp, *newfhp = NULL;
	char		*fname;
	int		nlen, err, type, mode;
	int		rdonly = 0, exists;
	dev_t		rdev = NODEV;

	DECODE_CREATEARGS(fhp, fname, nlen, attr);

	/* Get the directory inode */
	if (!(dirp = fh_inode(rqstp, fhp, S_IFDIR, MAY_EXEC, &err))) {
		fh_put(fhp);
		RETURN(err);
	}

	/* Check for MAY_WRITE separately. */
	err = nfsd_permission(rqstp, fhp->fh_export, dirp, MAY_WRITE);
	if (err == nfserr_rofs) {
		rdonly = 1;	/* Non-fatal error for echo > /dev/null */
	} else if (err) {
		fh_put(fhp);
		RETURN(err);
	}

	if (nfsd_iscovered(dirp)) {
		fh_put(fhp);
		RETURN(nfserr_acces);
	}

	/* First, check if the file already exists.  */
	exists = !nfsd_lookup(rqstp, fhp, fname, nlen, &newfhp);

	/* Unfudge the mode bits */
	if (attr.ia_valid & ATTR_MODE) { 
		type = attr.ia_mode & S_IFMT;
		mode = attr.ia_mode & ~S_IFMT;
		/* Weird. HP boxen seem to send a zero mode field at times */
		if (!type)
			type = S_IFREG;
	} else if (exists) {
		inode = newfhp->fh_inode;
		type = inode->i_mode & S_IFMT;
		mode = inode->i_mode & ~S_IFMT;
	} else {
		type = S_IFREG;
		mode = 0;	/* ??? */
	}

	/* This is for "echo > /dev/null" a la SunOS. Argh. */
	if (rdonly && (!exists || type == S_IFREG)) {
		err = nfserr_rofs;
		goto done;
	}

	attr.ia_valid |= ATTR_MODE;
	attr.ia_mode = type | mode;

	/* Handle unusual file types first */
	err = 0;
	if (type != S_IFREG) {
		int	is_borc = 0;
		u32	size = attr.ia_size;

		if (type != S_IFBLK && type != S_IFCHR) {
			rdev = 0;
		} else if (size & 0xffff0000) {
			/* Some OSs (e.g. HPUX) use a 32bit dev_t... Shrug */
			err = nfserr_io;
			goto done;
		} else {
			rdev = makedev((size >> 8) & 0xff, size & 0xff);

			/* If you think you've seen the worst, grok this. */
			if (type == S_IFCHR && rdev == 0xFFFF) {
				attr.ia_mode = S_IFIFO | mode;
				type = S_IFIFO;
			} else
				is_borc = 1;
		}

		if (!exists) {
			switch (type) {
			case S_IFCHR: case S_IFBLK:
			case S_IFIFO:
				err = nfsd_mknod(dirp, fname, nlen,
						attr.ia_mode, rdev);
				break;
			case S_IFSOCK:
				printk(KERN_NOTICE "nfsd: no unix sockets!\n");
				/* err = nfsd_unixbind(foo, bar, baz); */
				err = nfserr_perm;
				break;
			default:
				err = nfserr_acces;
			}
			if (!err) {
				err = nfsd_lookup(rqstp, fhp, fname, nlen,
						  &newfhp);
			}
		} else {
			/* Make sure the type and device matches */
			if (type != (inode->i_mode & S_IFMT)
			 || (is_borc && inode->i_rdev != rdev))
				err = nfserr_exist;
		}
	} else if (!exists) {
		err = nfsd_create(rqstp, fhp, fname, nlen, attr.ia_mode,
					&newfhp);
	}

	if (err)
		goto done;
	inode = newfhp->fh_inode;

	/* Set attributes only for non-existing files. This makes create
	 * behave like open(..., O_CREAT|O_TRUNC|O_WRONLY). Only the
	 * size attribute may be set to allow truncate() to work properly.
	 */
	if (!exists)
		attr.ia_valid &= ~(ATTR_SIZE);

	/* create() overloading for truncate. */
	if (S_ISREG(inode->i_mode) && (attr.ia_valid & ATTR_SIZE)
	 && attr.ia_size != inode->i_size) {
		err = nfsd_truncate(rqstp, inode, attr.ia_size);
		if (err)
			return err;
		attr.ia_valid |= ATTR_MTIME;
		attr.ia_mtime = CURRENT_TIME;
	}

	/* Avoid calling notify_change unless necessary. */
	if (!((attr.ia_mode ^ inode->i_mode) & S_IALLUGO))
		attr.ia_valid &= ~ATTR_MODE;
	if (attr.ia_uid == inode->i_uid)
		attr.ia_valid &= ~ATTR_UID;
	if (attr.ia_gid == inode->i_gid)
		attr.ia_valid &= ~ATTR_GID;
	if (attr.ia_valid) {
		err = nfsd_setattr(rqstp, newfhp, &attr);
		if (err) /* Should we call write_inode nvertheless? */
			goto done;
	}

	if (EX_ISSYNC(fhp->fh_export)) {
		write_inode(inode);
		write_inode(dirp);
	}

done:
	if (!err) {
		/* Encode the whole stuff. */
		if (!nfsd_encode_diropokres(rqstp, resp, newfhp))
			err = rpc_system_err;
		else
			err = rpc_success;
	} else {
		nfsd_putstat(resp, err);
		err = rpc_success;
	}

	if (newfhp)
		fh_put(newfhp);
	fh_put(fhp);

	return err;
}

static int
nfsd_proc_remove(svc_req *rqstp, svc_buf *argp, svc_buf *resp)
{
	svc_fh		*fhp;
	char		*fname;
	int		flen;
	int		err;

	DECODE_DIROPARGS(fhp, fname, flen);

	err = nfsd_remove(rqstp, fhp, fname, flen);

	fh_put(fhp);
	RETURN(err);
}

static int
nfsd_proc_rename(svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	svc_fh		*fhp, *tofhp;
	char		*fname, *tname;
	int		flen, tlen;
	int		err;

	DECODE_DIROPARGS(fhp, fname, flen);
	DECODE_DIROPARGS2(tofhp, tname, tlen, fhp);

	err = nfsd_rename(rqstp, fhp, fname, flen, tofhp, tname, tlen);

	fh_put(fhp);
	fh_put(tofhp);

	RETURN(err);
}

static int
nfsd_proc_link (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	svc_fh		*fhp, *tofhp;
	char		*fname;
	int		nlen, err;

	DECODE_FH(fhp);
	DECODE_DIROPARGS2(tofhp, fname, nlen, fhp);

	err = nfsd_link(rqstp, fhp, fname, nlen, tofhp);

	fh_put(tofhp);
	fh_put(fhp);

	RETURN(err);
}

static int
nfsd_proc_symlink (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	struct iattr	attr;
	svc_fh		*fhp, *newfhp;
	char		*fname, *sname;
	int		nlen, slen, err;

	DECODE_DIROPARGS(fhp, fname, nlen);

	/* Get symlink file name */
	if (!nfsd_decode_filename(rqstp, argp, &sname, &slen)) {
		fh_put(fhp);
		return rpc_garbage_args;
	}

	/* Get file attrs */
	if (!nfsd_decode_sattr(rqstp, argp, &attr)) {
		fh_put(fhp);
		return rpc_garbage_args;
	}

	/* Create the link, look up new file and set attrs.
	 * FIXME: should create newfh right in nfsd_symlink.
	 */
	if (!(err = nfsd_symlink(rqstp, fhp, fname, nlen, sname, slen)) &&
	    !(err = nfsd_lookup(rqstp, fhp, fname, nlen, &newfhp))) {
		err = nfsd_setattr(rqstp, newfhp, &attr);
		fh_put(newfhp);
	}

	fh_put(fhp);
	RETURN(err);
}

/*
 * Make directory. This operation is not idempotent.
 */
static int
nfsd_proc_mkdir (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	struct iattr	attr;
	svc_fh		*fhp, *newfhp;
	char		*fname;
	int		nlen, err;

	DECODE_CREATEARGS(fhp, fname, nlen, attr);

	if ((err = nfsd_mkdir(rqstp, fhp, fname, nlen, &attr, &newfhp)) != 0) {
		fh_put(fhp);
		RETURN(err);
	}

	/* Encode result. */
	if (!nfsd_encode_diropokres(rqstp, resp, newfhp))
		err = rpc_system_err;
	else
		err = rpc_success;

	fh_put(newfhp);
	fh_put(fhp);
	return err;
}

static int
nfsd_proc_rmdir (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	svc_fh		*fhp;
	char		*fname;
	int		nlen, err;

	DECODE_DIROPARGS(fhp, fname, nlen);

	err = nfsd_rmdir(rqstp, fhp, fname, nlen);

	fh_put(fhp);
	RETURN(err);
}

/*
 * Read a portion of a directory.
 */
static int
nfsd_proc_readdir (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	svc_fh		*fhp;
	u32		*temp;
	u32		cookie, size, rbuflen;
	int		err, eof, sizeq;

	DECODE_FH(fhp);

	if (argp->len < 2)
		return rpc_garbage_args;

	cookie = ntohl(*(argp->buf++));
	size = ntohl(*(argp->buf++));
	sizeq = size >> 2;

	temp = resp->buf;
	nfsd_putstat(resp, nfs_ok);

	/* make sure we have enough room. */
	if (resp->len + 2 > resp->buflen) {
		fh_put(fhp);
		return rpc_system_err;
	}
	resp->len += 2;	 /* reserve room for NULL ptr & eof marker */

	/* Hack: adjust buffer length to client read size */
	rbuflen = resp->buflen;
	if (resp->buflen > resp->len + sizeq)
		resp->buflen = resp->len + sizeq;

	eof = 1;
	err = nfsd_readdir(rqstp, fhp, (loff_t) cookie, 
				nfsd_encode_entry, resp, &eof);
	if (!err) {
		/* mark `no more entries' and set eof flag */
		*(resp->buf++) = 0;
		*(resp->buf++) = eof ? xdr_one : 0;
	} else {
		*temp++ = err;
		resp->len = temp - resp->base;
	}

	resp->buflen = rbuflen;	/* restore original buflen */

	fh_put(fhp);
	return rpc_success;
}

/*
 * Get file system info
 */
static int
nfsd_proc_statfs (svc_req * rqstp, svc_buf * argp, svc_buf * resp)
{
	struct super_block *sb;
	struct statfs	sbuf;
	struct inode	*inode;
	unsigned long	oldfs;
	svc_fh		*fhp;
	int		err;

	DECODE_FH(fhp);

	if (resp->len < 6)
		return rpc_system_err;

	/* FIXME: use (yet-to-be-done) nfsd_get_super instead */
	if (!(inode = fh_inode(rqstp, fhp, 0, MAY_NOP, &err))) {
		fh_put(fhp);
		RETURN(err);
	}

	/* Move to fileop.c */
	if (!(sb = inode->i_sb) || !sb->s_op->statfs) {
		fh_put(fhp);
		RETURN(nfserr_io);
	}

	oldfs = get_fs();
	set_fs (get_ds());
	sb->s_op->statfs (sb, &sbuf, sizeof(sbuf));
	set_fs (oldfs);

	nfsd_putstat (resp, nfs_ok);
	*(resp->buf++) = htonl (8 * 1024);
	*(resp->buf++) = htonl (sbuf.f_bsize);
	*(resp->buf++) = htonl (sbuf.f_blocks);
	*(resp->buf++) = htonl (sbuf.f_bfree);
	*(resp->buf++) = htonl (sbuf.f_bavail);
	resp->len += 5;

	fh_put(fhp);
	return rpc_success;
}

/* Map errnos to NFS errnos */
int
nfserrno (int errno)
{
	static struct {
		int	nfserr;
		int	syserr;
	} nfs_errtbl[] = {
		{ NFS_OK, 0 },
		{ NFSERR_PERM, EPERM },
		{ NFSERR_NOENT, ENOENT },
		{ NFSERR_IO, EIO },
		{ NFSERR_NXIO, ENXIO },
		{ NFSERR_ACCES, EACCES },
		{ NFSERR_EXIST, EEXIST },
		{ NFSERR_NODEV, ENODEV },
		{ NFSERR_NOTDIR, ENOTDIR },
		{ NFSERR_ISDIR, EISDIR },
		{ NFSERR_INVAL, EINVAL },
		{ NFSERR_FBIG, EFBIG },
		{ NFSERR_NOSPC, ENOSPC },
		{ NFSERR_ROFS, EROFS },
		{ NFSERR_NAMETOOLONG, ENAMETOOLONG },
		{ NFSERR_NOTEMPTY, ENOTEMPTY },
#ifdef EDQUOT
		{ NFSERR_DQUOT, EDQUOT },
#endif
		{ NFSERR_STALE, ESTALE },
		{ NFSERR_WFLUSH, EIO },
		{ -1, EIO }
	};
	int	i;

	for (i = 0; nfs_errtbl[i].nfserr != -1; i++) {
		if (nfs_errtbl[i].syserr == errno)
			return htonl (nfs_errtbl[i].nfserr);
	}
	printk (KERN_INFO "nfsd: non-standard errno: %d\n", errno);
	return nfserr_io;
}

#if 0
static void
nfsd_dump(char *tag, u32 *buf, int len)
{
	int	i;

	printk(KERN_NOTICE
		"nfsd: %s (%d words)\n", tag, len);

	for (i = 0; i < len && i < 32; i += 8)
		printk(KERN_NOTICE
			" %08lx %08lx %08lx %08lx"
			" %08lx %08lx %08lx %08lx\n",
			buf[i],   buf[i+1], buf[i+2], buf[i+3],
			buf[i+4], buf[i+5], buf[i+6], buf[i+7]);
}
#endif
