/*
 * nfssock.c	These are the socket receive routines for nfsd.
 *
 *		The server scheduling algorithm does not appear to
 *		distribute the load evenly, watching with top reveals
 *		a deviation of up to 20 percent from average.
 *
 *		TCP support is largely untested and may be a little
 *		slow. The problem is that we currently do two separate
 *		recvfrom's, one for the 4-byte record length, and the
 *		second for the actual record. This could possibly be
 *		improved by always reading a minimum size of around 
 *		100 bytes and tucking any superfluous bytes away in
 *		a temporary store. Still, that leaves write requests
 *		out in the rain. An alternative may be to peek at
 *		the first skb in the queue, and if it matches the next
 *		TCP sequence number, to extract the record marker.
 *		Yuck.
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#ifdef MODULE
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/linkage.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>

/*
 * We must get at the kernel-only network header files.
 * This is a gross hack.
 */
#include <linux/netdevice.h>
#include <linux/../net/sock.h>

#include "nfsd.h"

#ifdef DEBUG
#define dprintk(args...)	ifdebug(SOCK) printk(## args)
#else
#define dprintk(args...)	/* nothing */
#endif

static svc_sock *nfsd_setup_socket(struct socket *sock, svc_sock **list);
static void	nfsd_remove_socket(svc_sock *ssk);
static void	nfsd_udp_data_ready(struct sock *sk, int count);
static int	nfsd_udp_recvfrom(struct svc_req *rqstp);
static int	nfsd_udp_sendto(svc_req *rqstp);

typedef struct svc_item {
	struct svc_item	*prev;
	struct svc_item	*next;
} svc_item;

typedef struct svc_list {
	svc_item	*head;
	svc_item	*tail;
} svc_list;

static svc_sock *	allsocks = NULL;
static svc_sock *	tcpsock = NULL;
static svc_list		servers = { NULL, NULL };
static svc_list		sockets = { NULL, NULL };

/*
 * Add a process to the queue of rqhead.
 */
static inline void
nfsd_append(svc_list *list, svc_item *what)
{
	svc_item	*prev;

	if ((prev = list->tail) != NULL)
		prev->next = what;
	else
		list->head = what;
	what->prev = prev;
	what->next = NULL;
	list->tail = what;
}

static inline void
nfsd_unlink(svc_list *list, svc_item *what)
{
	svc_item	*prev = what->prev,
			*next = what->next;

	if (prev != NULL)
		prev->next = next;
	else
		list->head = next;
	if (next != NULL)
		next->prev = prev;
	else
		list->tail = prev;
	what->prev = NULL;
	what->next = NULL;
}

/*
 * Queue up an idle nfsd thread.
 */
static inline void
nfsd_serv_enqueue(struct svc_req *rqstp)
{
	nfsd_append(&servers, (svc_item *) rqstp);
}

/*
 * Dequeue an nfsd thread.
 */
static inline void
nfsd_serv_dequeue(svc_req *rqstp)
{
	nfsd_unlink(&servers, (svc_item *) rqstp);
}

/*
 * Queue up a socket with data pending. If there are idle nfsd
 * processes, wake'em up.
 */
static inline void
nfsd_sock_enqueue(struct svc_sock *ssk)
{
	svc_item	*head;
	svc_req		*rqstp;

	if (ssk->sk_busy) {
		/* Don't enqueue socket while daemon is receiving */
		dprintk("socket %p not enqueued: busy\n", ssk->sk_sk);
		return;
	}
	ssk->sk_busy = 1;

	if ((head = servers.head) != NULL) {
		dprintk("socket %p served by daemon %p\n", ssk->sk_sk, head);
		nfsd_unlink(&servers, head);
		rqstp = (svc_req *) head;
		rqstp->rq_sock = ssk;
		ssk->sk_inuse++;
		wake_up(&rqstp->rq_wait);
	} else {
		dprintk("socket %p put into queue\n", ssk->sk_sk);
		nfsd_append(&sockets, (svc_item *) ssk);
	}
}

/*
 * Dequeue the first socket.
 */
static inline svc_sock *
nfsd_sock_dequeue(void)
{
	unsigned long	oldflags;
	svc_item	*head;

	save_flags(oldflags); cli();
	if ((head = sockets.head) != NULL)
		nfsd_unlink(&sockets, head);

	if (head) {
		dprintk("socket %p dequeued\n", ((svc_sock *)head)->sk_sk);
	}
	restore_flags(oldflags);
	return (svc_sock *) head;
}

/*
 * Having read count bytes from a socket, check whether it
 * needs to be re-enqueued.
 */
static inline void
nfsd_sock_received(struct svc_sock *ssk, int count)
{
	unsigned long	oldflags;

	save_flags(oldflags); cli();
	if ((ssk->sk_data -= count) < 0) {
		printk(KERN_NOTICE "nfsd: sk_data negative!\n");
		ssk->sk_data = 0;
	}
	ssk->sk_rqstp = NULL; /* XXX */
	ssk->sk_busy = 0;
	if (ssk->sk_conn || ssk->sk_data || ssk->sk_close) {
		dprintk("socket %p re-enqueued after receive\n", ssk->sk_sk);
		nfsd_sock_enqueue(ssk);
	}
	restore_flags(oldflags);
}

/*
 * Dequeue a new connection.
 */
static inline void
nfsd_sock_accepted(struct svc_sock *ssk)
{
	unsigned long	oldflags;

	save_flags(oldflags); cli();
        ssk->sk_busy = 0;
        ssk->sk_conn--;
        if (ssk->sk_conn || ssk->sk_data || ssk->sk_close) {
                dprintk("socket %p re-enqueued after accept\n", ssk->sk_sk);
                nfsd_sock_enqueue(ssk);
        }
	restore_flags(oldflags);
}

/*
 * Release a socket after use.
 */
static inline void
nfsd_sock_release(svc_req *rqstp)
{
	svc_sock	*ssk = rqstp->rq_sock;

	if (!ssk)
		return;
	rqstp->rq_sock = NULL;
	ssk->sk_inuse--;
}

/*
 * INET callback when data has been received on the socket.
 */
static void
nfsd_udp_data_ready(struct sock *sk, int count)
{
	svc_sock *	ssk;

	dprintk("socket %p data ready (inet %p)\n", sk->user_data, sk);
	ssk = (svc_sock *)(sk->user_data);
	if (!ssk)
		return;
	ssk->sk_data++;
	nfsd_sock_enqueue(ssk);
}

/*
 * Receive a datagram from a UDP socket.
 */
static int
nfsd_udp_recvfrom(struct svc_req *rqstp)
{
	unsigned long	oldfs;
	svc_sock	*ssk = rqstp->rq_sock;
	struct socket	*sock = ssk->sk_sock;
	struct iovec	iov;
	struct msghdr	msg;
	int		len, ready;

	rqstp->rq_addrlen = sizeof(rqstp->rq_addr);

	msg.msg_name = &rqstp->rq_addr;
	msg.msg_namelen = sizeof(rqstp->rq_addr);
	msg.msg_iov  = &iov;
	msg.msg_iovlen = 1;
	msg.msg_accrights = 0;
	iov.iov_base = rqstp->rq_args.base;
	iov.iov_len  = rqstp->rq_args.buflen << 2;
	ready = ssk->sk_data;

	dprintk("socket %p UDP calling recvfrom(%p, %d)\n",
			sock->data, iov.iov_base, iov.iov_len);
	oldfs = get_fs();
	set_fs(get_ds());
	len = sock->ops->recvmsg(sock, &msg, iov.iov_len, 1, 0,
			&rqstp->rq_addrlen);
	set_fs(oldfs);

	dprintk("socket %p return value %d\n", sock->data, len);

	if (len > 0) {
		rqstp->rq_args.buf = rqstp->rq_args.base;
		rqstp->rq_args.len = (len >> 2);
		rqstp->rq_res.buf = rqstp->rq_res.base;
		rqstp->rq_res.len = 0;
		nfsd_sock_received(ssk, 1);
		nfsdstats.netudpcnt++;
	} else if (len == -EAGAIN) {
		printk(KERN_NOTICE
		      "nfsd: weird: %d packets ready, but got EAGAIN\n", ready);
#if 0
		nfsd_sock_received(ssk, ready); /* Clear data ready */
#else
		nfsd_sock_received(ssk, 1);
#endif
	} else {
		dprintk(KERN_NOTICE "nfsd: recvfrom returned errno %d\n", -len);
		nfsd_sock_received(ssk, 0);
	}

	return len;
}

static int
nfsd_udp_sendto(svc_req *rqstp)
{
	unsigned long	oldfs = get_fs();
	struct socket	*sock = rqstp->rq_sock->sk_sock;
	struct msghdr	msg;
	struct iovec	iov;
	int		len;

	msg.msg_name = &rqstp->rq_addr;
	msg.msg_namelen = sizeof(rqstp->rq_addr);
	msg.msg_iov  = &iov;
	msg.msg_iovlen = 1;
	msg.msg_accrights = 0;
	iov.iov_base = rqstp->rq_res.base;
	iov.iov_len  = rqstp->rq_res.len << 2;

#if 0
	{
	u32		*bufp = rqstp->rq_res.base;
        printk("reply len %d\n", iov.iov_len);
        printk("0000 %08x %08x %08x %08x %08x %08x %08x %08x\n",
                        bufp[0], bufp[1], bufp[2], bufp[3],
                        bufp[4], bufp[5], bufp[6], bufp[7]);
        printk("0020 %08x %08x %08x %08x %08x %08x %08x %08x\n",
                        bufp[8],  bufp[9],  bufp[10], bufp[11],
                        bufp[12], bufp[13], bufp[14], bufp[15]);
	}
#endif

	set_fs(get_ds());
	len = sock->ops->sendmsg(sock, &msg, iov.iov_len, 1, 0);
	set_fs(oldfs);

	return len;
}

static int
nfsd_udp_init(svc_sock *ssk)
{
	ssk->sk_sk->data_ready = nfsd_udp_data_ready;
	ssk->sk_recvfrom = nfsd_udp_recvfrom;
	ssk->sk_sendto = nfsd_udp_sendto;

	return 0;
}

/*
 * A state change on a listening socket means there's a connection
 * pending.
 */
static void
nfsd_tcp_state_change1(struct sock *sk)
{
	dprintk("socket %p TCP (listen) state change %d\n",
			sk, sk->state);

	if  (sk->state != TCP_ESTABLISHED) {
		/* Aborted connection, SYN_RECV or whatever... */
		return;
	}
	tcpsock->sk_conn++;
	nfsd_sock_enqueue(tcpsock);
}

/* A state change on a connected socket means it's dying or dead */
static void
nfsd_tcp_state_change2(struct sock *sk)
{
	svc_sock	*ssk;

	dprintk("socket %p TCP (connected) state change %d (ssk %p)\n",
			sk, sk->state, sk->user_data);

	if (!(ssk = (svc_sock *) sk->user_data)) {
		printk("socket %p: no user data\n", sk);
		return;
	}
	ssk->sk_close++;
	nfsd_sock_enqueue(ssk);
}

static void
nfsd_tcp_data_ready(struct sock *sk, int count)
{
	svc_sock *	ssk;

	/* Disconnect signalled through data_ready?!? */
	if (sk->state != TCP_ESTABLISHED) {
		nfsd_tcp_state_change2(sk);
		return;
	}

	dprintk("socket %p TCP data ready (ssk %p)\n",
			sk, sk->user_data);
	if (!(ssk = (svc_sock *)(sk->user_data)))
		return;
	ssk->sk_data++;
	nfsd_sock_enqueue(ssk);
}

/*
 * Accept a TCP connection
 */
static void
nfsd_tcp_accept(svc_sock *ssk)
{
	struct sockaddr_in sin;
	struct socket	*sock = ssk->sk_sock;
	struct socket	*newsock;
	struct proto_ops *ops;
	svc_client	*client;
	svc_sock	*newssk;
	int		err, slen;

	dprintk("tcp_accept %p sock %p\n", ssk, sock);
	if (!sock)
		return;

	if (!(newsock = sock_alloc())) {
		printk(KERN_WARNING "nfsd: no more sockets!\n");
		return;
	}
	dprintk("tcp_accept %p allocated\n", newsock);

	newsock->type = sock->type;
	newsock->ops = ops = sock->ops;
#if 0
	if ((err = sock->ops->dup(newsock, sock)) < 0) {
		printk(KERN_WARNING "nfsd: dup failed (err %d)!\n", -err);
		goto failed;
	}
#endif

	if ((err = ops->accept(sock, newsock, O_NONBLOCK)) < 0) {
		printk(KERN_WARNING "nfsd: accept failed (err %d)!\n", -err);
		goto failed;		/* aborted connection or whatever */
	}

	slen = sizeof(sin);
	err = ops->getname(newsock, (struct sockaddr *) &sin, &slen, 1);
	if (err < 0) {
		printk(KERN_WARNING "nfsd: peername failed (err %d)!\n", -err);
		goto failed;		/* aborted connection or whatever */
	}

	/* Verify the client's address */
	exp_readlock();
	client = exp_getclient(&sin.sin_addr);
	exp_unlock();

	if (client == NULL) {
		printk(KERN_WARNING "nfsd: rejected connect from %08lx:%04x\n",
			ntohl(sin.sin_addr.s_addr), ntohs(sin.sin_port));
		goto failed;
	}

	printk(KERN_DEBUG "nfsd: connect from %08lx:%04x (%s)\n",
			ntohl(sin.sin_addr.s_addr), ntohs(sin.sin_port),
			client->cl_ident);

	if (!(newssk = nfsd_setup_socket(newsock, &allsocks))) {
		goto failed;
	}

	/* Precharge. Data may have arrived on the socket before we
	 * installed the data_Ready callback. 
	 */
	newssk->sk_data = 1;
	newssk->sk_temp = 1;
	nfsd_sock_enqueue(newssk);

	nfsdstats.nettcpconn++;

	return;

failed:
	sock_release(newsock);
	return;
}

/*
 * Receive data from a TCP socket.
 */
static int
nfsd_tcp_recvfrom(struct svc_req *rqstp)
{
	unsigned long	oldfs = get_fs();
	svc_sock	*ssk = rqstp->rq_sock;
	struct socket	*sock = ssk->sk_sock;
	struct sock	*sk;
	struct iovec	iov;
	struct msghdr	msg;
	u32		*buffer, reclen;
	int		buflen, len, ready;

	dprintk("tcp_recv %p data %d conn %d close %d\n",
			ssk, ssk->sk_data, ssk->sk_conn, ssk->sk_close);
	if (ssk->sk_close) {
		sk = ssk->sk_sk;
		nfsd_remove_socket(ssk);
		return 0;
	}

	if (ssk->sk_conn) {
		nfsd_tcp_accept(ssk);
		nfsd_sock_accepted(ssk);
		return 0;
	}

	rqstp->rq_addrlen = sizeof(rqstp->rq_addr);
	ready = ssk->sk_data;

	while (1) {
		u8	*buffer;

		buffer = (u8 *) ssk->sk_rbuf + ssk->sk_rlen;
		buflen = ssk->sk_reclen + 4 - ssk->sk_rlen;

		dprintk("socket %p TCP calling recvfrom(%p, %d)\n",
				sock->data, buffer, buflen);
		msg.msg_name = &rqstp->rq_addr;
		msg.msg_namelen = sizeof(rqstp->rq_addr);
		msg.msg_iov  = &iov;
		msg.msg_iovlen = 1;
		msg.msg_accrights = 0;
		iov.iov_base = buffer;
		iov.iov_len  = buflen;
		set_fs(get_ds());
		len = sock->ops->recvmsg(sock, &msg, buflen, 1, 0,
				&rqstp->rq_addrlen);
		set_fs(oldfs);
		dprintk("TCP recvfrom return value %d\n", len);

		if (len < 0)
			break;
		ssk->sk_rlen += len;

		if (ssk->sk_reclen || len < buflen)
			break;
		reclen = ntohl(ssk->sk_rbuf[0]);
		if (!(reclen & 0x80000000)) {
			/* FIXME: shutdown socket */
			printk("bad reclen %08lx", (unsigned long) reclen);
			return -EIO;
		}
		ssk->sk_reclen = reclen & 0x7fffffff;
		dprintk("TCP record, %ld bytes\n", ssk->sk_reclen);
	}


	if (len > 0) {
		len = ssk->sk_rlen - 4;
		if (ssk->sk_reclen && len < ssk->sk_reclen) {
			dprintk("got incomplete record (%d of %ld)\n",
					len, ssk->sk_reclen);
			nfsd_sock_received(ssk, ready);
			len = -EAGAIN;	/* record not complete */
		} else {
			dprintk("got complete record (%d bytes)\n", len);

			/* swap argument buffer and TCP buffer */
			buffer = rqstp->rq_args.base;
			rqstp->rq_args.base = ssk->sk_rbuf;
			rqstp->rq_args.buf = ssk->sk_rbuf + 1;
			rqstp->rq_args.len = (len >> 2);

			/* Set up reply buffer */
			rqstp->rq_res.buf = rqstp->rq_res.base + 1;
			rqstp->rq_res.len = 0;

			ssk->sk_rbuf = buffer;
			ssk->sk_reclen = 0;
			ssk->sk_rlen = 0;

			nfsd_sock_received(ssk, 1);
			nfsdstats.nettcpcnt++;
		}
	} else if (len == -EAGAIN) {
		dprintk("knfsd: TCP recvfrom got EAGAIN\n");
		nfsd_sock_received(ssk, ready); /* Clear data ready */
	} else {
		printk(KERN_NOTICE "nfsd: recvfrom returned errno %d\n", -len);
		nfsd_sock_received(ssk, 0);
	}

	return len;
}

/*
 * Send out data on TCP socket.
 * FIXME: Make the sendto call non-blocking in order not to hang
 * all nfsds ona a dead client. Requires write queue maintenance.
 */
static int
nfsd_tcp_sendto(svc_req *rqstp)
{
	unsigned long	oldfs = get_fs();
	struct socket	*sock = rqstp->rq_sock->sk_sock;
	struct iovec	iov;
	struct msghdr	msg;
	u32		*buffer = rqstp->rq_res.base;
	int		buflen = rqstp->rq_res.len << 2, len;

	msg.msg_name = &rqstp->rq_addr;
	msg.msg_namelen = sizeof(rqstp->rq_addr);
	msg.msg_iov  = &iov;
	msg.msg_iovlen = 1;
	msg.msg_accrights = 0;
	iov.iov_base = buffer;
	iov.iov_len  = buflen + 4;

	buffer[0] = htonl(0x80000000|buflen);

	set_fs(get_ds());
	len = sock->ops->sendmsg(sock, &msg, buflen + 4, 0, 0);
	set_fs(oldfs);

	return len;
}

static int
nfsd_tcp_init(svc_sock *ssk)
{
	struct sock	*sk = ssk->sk_sk;

	ssk->sk_recvfrom = nfsd_tcp_recvfrom;
	ssk->sk_sendto = nfsd_tcp_sendto;

	if (sk->state == TCP_LISTEN) {
		dprintk("setting up TCP socket for listening\n");
		sk->state_change = nfsd_tcp_state_change1;
		ssk->sk_rbuf = NULL;
	} else {
		dprintk("setting up TCP socket for reading\n");
		sk->state_change = nfsd_tcp_state_change2;
		sk->data_ready = nfsd_tcp_data_ready;

		if (!(ssk->sk_rbuf = (u32 *) kmalloc(MAXBUF, GFP_NFS)))
			return -ENOMEM;
		ssk->sk_reclen = 0;
		ssk->sk_rlen = 0;
	}

	return 0;
}

/*
 * Receive the next request on any socket.
 */
int
nfsd_recv(svc_req *rqstp)
{
	unsigned long	oldflags;
	svc_sock *	ssk;
	int		len;

	if (!current->timeout) {
		dprintk("server %p waiting for data\n", rqstp);
	}
again:
	save_flags(oldflags); cli();
	if ((ssk = nfsd_sock_dequeue()) != NULL) {
		rqstp->rq_sock = ssk;
		ssk->sk_inuse++;
		restore_flags(oldflags);
	} else {
		/* No data pending. Go to sleep */
		rqstp->rq_sock = NULL;
		rqstp->rq_wait = NULL;
		nfsd_serv_enqueue(rqstp);

		interruptible_sleep_on(&rqstp->rq_wait);
		restore_flags(oldflags);

		if (!(ssk = rqstp->rq_sock)) {
			nfsd_serv_dequeue(rqstp);
			return -EAGAIN;
		}
	}

	len = ssk->sk_recvfrom(rqstp);
	if (len == 0) {	/* No data, incomplete (TCP) read, or accept() */
		nfsd_sock_release(rqstp);
		goto again;
	}

	nfsdstats.netcnt++;
	return len;
}

/* 
 * Drop request
 */
void
nfsd_drop(svc_req *rqstp)
{
	dprintk("socket %p dropped request\n", rqstp->rq_sock);
	nfsd_sock_release(rqstp);
}

/*
 * Return reply to client.
 */
int
nfsd_send(svc_req *rqstp)
{
	svc_sock	*ssk;
	int		len;

	if ((ssk = rqstp->rq_sock) == NULL) {
		printk(KERN_WARNING "NULL socket pointer in %s:%d\n",
				__FILE__, __LINE__);
		return -EFAULT;
	}
	len = ssk->sk_sendto(rqstp);
	nfsd_sock_release(rqstp);
	return len;
}

static svc_sock *
nfsd_setup_socket(struct socket *sock, svc_sock **list)
{
	svc_sock	*ssk;
	struct sock	*sk;
	int		err;

	if (!(ssk = kmalloc(sizeof(*ssk), GFP_ATOMIC)))
		return NULL;
	memset(ssk, 0, sizeof(*ssk));

	sk = (struct sock *) sock->data;
	sk->user_data = ssk;
	ssk->sk_sock = sock;
	ssk->sk_sk = sk;
	ssk->sk_ostate = sk->state_change;
	ssk->sk_odata = sk->data_ready;

	if (sock->type == SOCK_DGRAM)
		err = nfsd_udp_init(ssk);
	else
		err = nfsd_tcp_init(ssk);
	if (err < 0) {
		sk->user_data = NULL;
		kfree(ssk);
		return NULL;
	}

	ssk->sk_list = *list;
	*list = ssk;

	return ssk;
}

int
nfsd_add_socket(struct socket *sock)
{
	svc_sock	*ssk;
	svc_sock	**socklist;

	if (sock->ops->family != AF_INET) {
		printk(KERN_WARNING "nfsd: only INET sockets supported\n");
		return -EBADF;
	}
	if (sock->type != SOCK_DGRAM && sock->type != SOCK_STREAM) {
		printk(KERN_WARNING "nfsd: only UDP and TCP "
				"sockets supported\n");
		return -EBADF;
	}

	if (sock->type == SOCK_STREAM) {
		if (sock->state != SS_UNCONNECTED ||
		  !(sock->flags & SO_ACCEPTCON)) {
			printk(KERN_WARNING
				"nfsd: socket not in LISTEN state\n");
			return -EINVAL;
		}
		socklist = &tcpsock;
	} else {
		socklist = &allsocks;
	}

	for (ssk = *socklist; ssk; ssk = ssk->sk_list) {
		if (ssk->sk_sock == sock)
			return 0;
	}

	if (!(ssk = nfsd_setup_socket(sock, socklist)))
		return -ENOMEM;

	return 0;
}

static void
nfsd_remove_socket(svc_sock *ssk)
{
	svc_sock	**rsk;

	if (!ssk->sk_temp) {
		printk(KERN_WARNING "nfsd: removing server socket!\n");
		return;
	}

	for (rsk = &allsocks; *rsk; rsk = &(*rsk)->sk_list) {
		if (*rsk == ssk)
			break;
	}

	if (!*rsk)
		return;
	*rsk = ssk->sk_next;

	sock_release(ssk->sk_sock);
	kfree(ssk);
}

/*
 * The following must be called when shutting down the module.
 */
static void
nfsd_remove_socket_list(svc_sock **head)
{
	svc_sock	*ssk;

	while ((ssk = *head) != NULL) {
		*head = ssk->sk_list;
		if (ssk->sk_temp) {
			/* Destroy socket */
			dprintk("Releasing socket %p\n", ssk->sk_sk);
			sock_release(ssk->sk_sock);
		} else {
			/* Restore default callbacks */
			dprintk("Restoring socket %p\n", ssk->sk_sk);
			ssk->sk_sk->state_change = ssk->sk_ostate;
			ssk->sk_sk->data_ready = ssk->sk_odata;
		}
		kfree(ssk);
	}
}

void
nfsd_shutdown_sockets(void)
{
	dprintk("shutting down socket module\n");
	sockets.head = sockets.tail = NULL;
	nfsd_remove_socket_list(&allsocks);
	nfsd_remove_socket_list(&tcpsock);
}
