/*
 * nfssvc.c	Central processing for nfsd.
 *
 * Authors:	Olaf Kirch (okir@monad.swb.de)
 *
 *		Inspired by a UDP echo service by Thomas Koenig
 *		(Thomas.Koenig@ciw.uni-karlsruhe.de)
 *
 * I'm doing quite a lot of set_fs(get_ds()) in knfsd. Maybe I could
 * just do a set_fs at the head of sys_nfsd_svc(), and reset it upon
 * exit. Not sure what that might break, though:-)
 *
 * FIXME: make knfsd a kernel thread.
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#ifdef MODULE
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/linkage.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/nfs.h>
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/malloc.h>
#include <linux/timer.h>
#include <netinet/in.h>	/* for IPPORT_RESERVED */

/*
 * We must get at the kernel-only network header files. 
 * This is a gross hack.
 */
#include <linux/netdevice.h>
#include <linux/../net/sock.h>

#include "nfsd.h"
#include <nfs/cache.h>

static int	nfsd_decode_auth (svc_req *rqstp, svc_buf *argp, svc_auth *ahp);
u32 *		nfsd_accept_reply(svc_req *rqstp, u32 rpc_stat);
void		nfsd_reject_version(svc_req *rqstp);
void		nfsd_reject_auth(svc_req *rqstp, u32 auth_stat);

#define BLOCKABLE_SIGS	(~((1 << (SIGKILL - 1)) | (1 << (SIGSTOP - 1))))
#define SIGALRM_MASK	(1 << (SIGALRM - 1))
#define FLUSH_INTVL	(NFSD_CACHE_EXPIRE/2)

/*
 * Version table for NFS
 */
struct svc_version	nfsd_version[NFSD_MAXVERS+1] = {
	{ NULL,		0 },
	{ NULL,		0 },
	{ nfsd_proc2,	(sizeof(nfsd_proc2) / sizeof(struct svc_proc)) },
#ifdef NFS3_SUPPORTED
	{ nfsd_proc3,	(sizeof(nfsd_proc3) / sizeof(struct svc_proc)) },
#endif
};

#ifdef DEBUG
#define dprintk(args...)	ifdebug(SVC) printk(## args)
#else
#define dprintk(args...)	/* nop */
#endif

/*
 * Miscellaneous pre-encoded RPC stat values.
 */
u32	rpc_success, rpc_prog_unavail, rpc_prog_mismatch, rpc_proc_unavail,
	rpc_garbage_args, rpc_system_err;
u32	auth_ok, autherr_badcred, autherr_rejectedcred, autherr_badverf,
	autherr_rejectedverf, autherr_tooweak;
u32	nfs_ok, nfserr_perm, nfserr_noent, nfserr_io, nfserr_nxio,
	nfserr_acces, nfserr_exist, nfserr_nodev, nfserr_notdir,
	nfserr_isdir, nfserr_fbig, nfserr_nospc, nfserr_rofs,
	nfserr_nametoolong, nfserr_dquot, nfserr_stale;
u32	xdr_zero, xdr_one, xdr_two;

/*
 * Get the next NFS request. In order to flush unused file handles
 * regularly, one of the daemons is put into `idle mode,' i.e. it will
 * return from nfsd_recv after FLUSH_INTVL jiffies.
 */
static inline int
nfsd_nextreq(struct svc_req *rqstp)
{
	static unsigned long	lastflush = 0;
	static int		idle = 0;
	int			idling = 0, len;

	if (!idle)
		idle = idling = 1;

	do {
		if (jiffies >= lastflush + FLUSH_INTVL) {
			dprintk("flushing... (idle %d)\n", idling);
			fh_flush(0);
			lastflush = jiffies;
		}

		if (idling)
			current->timeout = lastflush + FLUSH_INTVL;
		len = nfsd_recv(rqstp);
		current->timeout = 0;
	} while ((len == -EAGAIN || len == -ECONNREFUSED)
		 && !(current->signal & ~current->blocked));

	if (idling)
		idle = 0;

	return len;
}

/*
 * This is the toplevel service loop.
 */
int
nfsd_svc(int fds[], int nrfds)
{
	static int	nrservs = 0;	/* # of daemon processes */
	struct file	*file;		/* VFS layer */
	struct socket	*sock;		/* BSD socket layer */
	struct svc_req	*rqstp = NULL;
	int		len;
	unsigned long	oldfs;
	u32		*inbuf = NULL,
			*outbuf = NULL;
	int		oldumask;
	int		i, fd, err;

	/*
	 * Save state.
	 */
	oldumask = current->fs->umask;
	oldfs = get_fs();

	/*
	 * Register all sockets passed by nfsd.
	 */
	for (i = 0; i < nrfds; i++) {
		if ((fd = fds[i]) < 0 || fd >= NR_OPEN ||
		    !(file = current->files->fd[fd])) {
			err = -EBADF;
			goto finish0;
		}
		if (!file->f_inode || !file->f_inode->i_sock) {
			err = -ENOTSOCK;
			goto finish0;
		}
		sock = &file->f_inode->u.socket_i;
		nfsd_add_socket(sock);
	}
	nrservs++;

	/*
	 * Set umask to 0.
	 */
	current->fs->umask = 0;

	inbuf = kmalloc(MAXBUF, GFP_KERNEL);
	outbuf = kmalloc(MAXBUF, GFP_KERNEL);
	rqstp = kmalloc(sizeof(*rqstp), GFP_KERNEL);
	if (inbuf == NULL || outbuf == NULL || rqstp == NULL) {
		err = -ENOMEM;
		goto finish;
	}

	/*
	 * These should be moved to nfsrpc.c
	 */
	xdr_zero = htonl(0);
	xdr_one = htonl(1);
	xdr_two = htonl(2);

	rpc_success = htonl(RPC_SUCCESS);
	rpc_prog_unavail = htonl(RPC_PROG_UNAVAIL);
	rpc_prog_mismatch = htonl(RPC_PROG_MISMATCH);
	rpc_proc_unavail = htonl(RPC_PROC_UNAVAIL);
	rpc_garbage_args = htonl(RPC_GARBAGE_ARGS);
	rpc_system_err = htonl(RPC_SYSTEM_ERR);

	auth_ok = htonl(AUTH_OK);
	autherr_badcred = htonl(AUTH_BADCRED);
	autherr_rejectedcred = htonl(AUTH_REJECTEDCRED);
	autherr_badverf = htonl(AUTH_BADVERF);
	autherr_rejectedverf = htonl(AUTH_REJECTEDVERF);
	autherr_tooweak = htonl(AUTH_TOOWEAK);

	nfs_ok = htonl(NFS_OK);
	nfserr_perm = htonl(NFSERR_PERM);
	nfserr_noent = htonl(NFSERR_NOENT);
	nfserr_io = htonl(NFSERR_IO);
	nfserr_nxio = htonl(NFSERR_NXIO);
	nfserr_acces = htonl(NFSERR_ACCES);
	nfserr_exist = htonl(NFSERR_EXIST);
	nfserr_nodev = htonl(NFSERR_NODEV);
	nfserr_notdir = htonl(NFSERR_NOTDIR);
	nfserr_isdir = htonl(NFSERR_ISDIR);
	nfserr_fbig = htonl(NFSERR_FBIG);
	nfserr_nospc = htonl(NFSERR_NOSPC);
	nfserr_rofs = htonl(NFSERR_ROFS);
	nfserr_nametoolong = htonl(NFSERR_NAMETOOLONG);
	nfserr_dquot = htonl(NFSERR_DQUOT);
	nfserr_stale = htonl(NFSERR_STALE);

	rqstp->rq_args.base = inbuf;
	rqstp->rq_args.buflen = MAXBUF >> 2;
	rqstp->rq_res.base = outbuf;
	rqstp->rq_res.buflen = MAXBUF >> 2;

	while(1) {
		sigset_t	oldmask;
		int		dropit;

		/* Find a socket with data available and call its
		 * recvfrom routine.
		 */
		len = nfsd_nextreq(rqstp);
		CHECKPOINT();

		if (current->signal & ~current->blocked) {
			err = -EINTR;
			break;
		}

		if (len < 0) {
			printk(KERN_NOTICE
				"nfsd: terminating on recv error %d\n", -len);
			err = len;
			break;
		}

		/* Lock the export hash tables for reading. */
		exp_readlock();

		/* Validate the client's address. This will also defeat
		 * port probes on port 2049 by unauthorized clients.
		 */
		rqstp->rq_client = exp_getclient(&rqstp->rq_addr.sin_addr);
		if (!rqstp->rq_client) {
			printk(KERN_WARNING "nfsd: unauthenticated request "
				"from (%08lx:%04x)\n",
				ntohl(rqstp->rq_addr.sin_addr.s_addr),
				ntohs(rqstp->rq_addr.sin_port));
			nfsd_drop(rqstp);
			exp_unlock();
			nfsdstats.rpcbadclnt++;
			continue;	/* just drop it */
		}
		rqstp->rq_secure = ntohs(rqstp->rq_addr.sin_port)
						< IPPORT_RESERVED;
		rqstp->rq_userset = 0;

		/* Copy XID to reply buffer */
		xdr_getlong(&rqstp->rq_args, rqstp->rq_xid);
		xdr_putlong(&rqstp->rq_res, rqstp->rq_xid);

		/* Zero cache info */
		rqstp->rq_cacherep = NULL;

		/* safeguard */
		rqstp->rq_res.base[(MAXBUF >> 2) - 1] = 0xf00baa;

		/* Process request with all signals blocked.  */
		oldmask = current->blocked;
		current->blocked = BLOCKABLE_SIGS;
		dropit = !nfsd_process_req(rqstp);
		current->blocked = oldmask;

		/* Unlock export hash tables */
		exp_unlock();

		if (rqstp->rq_res.base[(MAXBUF >> 2) - 1] != 0xf00baa) {
			printk(KERN_WARNING "nfsd: written past eob!\n");
			nfsd_drop(rqstp);
			err = -EFAULT;
			break;
		}

		CHECKPOINT();

		if (dropit) {
			nfsd_drop(rqstp);
			continue;
		}

		/* return reply to client, finally! */
		len = nfsd_send(rqstp);
		CHECKPOINT();

		/*
		 * Connection refused means we got an ICMP_PORTUNREACH, possibly
		 * from an earlier sendto operation.
		 */
		if (len == -ECONNREFUSED || len == -EAGAIN)
			continue;

		if (len < 0) {
			if (len != -EINTR)
				printk(KERN_WARNING
					"nfsd: terminating on send error %d\n",
					-len);
			err = len;
			break;
		}
	}

	/* Flush the FH cache */
	fh_flush(1);

finish:
	/* This is currently done by the caller: */
	/* close_fp(file); */

finish0:
	if (--nrservs <= 0)
		nfsd_shutdown_sockets();
	if (inbuf)
		kfree(inbuf);
	if (outbuf)
		kfree(outbuf);
	if (rqstp)
		kfree(rqstp);
	current->fs->umask = oldumask;

	return err;
}

/*
 * Validate authentication info.
 */
static inline int
nfsd_validate_auth(svc_req *rqstp, u32 *errp)
{
	svc_authops *ops = rqstp->rq_cred.ah_op;

	return !ops->validate || ops->validate(rqstp, errp);
}

/*
 * Free authentication data.
 */
static inline void
nfsd_free_auth(svc_auth *ahp)
{
	svc_authops *ops = ahp->ah_op;

	if (ops->free)
		ops->free(ahp);
}

/*
 * Process the NFS request.
 */
int
nfsd_process_req (svc_req * rqstp)
{
	struct svc_version	*vp;
	struct svc_proc		*pp = NULL;	/* compiler food */
	svc_buf *		argp = &rqstp->rq_args;
	svc_buf *		resp = &rqstp->rq_res;
	u32			*bufp, *statp;
	u32			vers, proc, auth_stat, rpc_stat;
	unsigned long		t0 = jiffies;

	rpc_stat = rpc_success;
	bufp = argp->buf;
	rqstp->rq_cacherep = NULL;

	if (argp->len < 5)
		return 0;

/*
	printk("request\n");
	printk("0000 %08x %08x %08x %08x %08x %08x %08x %08x\n",
			rqstp->rq_xid,
			          bufp[0],  bufp[1],  bufp[2],
			bufp[3],  bufp[4],  bufp[5],  bufp[6]);
	printk("0020 %08x %08x %08x %08x %08x %08x %08x %08x\n",
			bufp[7],  bufp[8],  bufp[9],  bufp[10],
			bufp[11], bufp[12], bufp[13], bufp[14]);
 */

	if (ntohl(*bufp++) != 0) {		/* direction != CALL */
		nfsdstats.rpcbadfmt++;
		return 0;			/* drop it */
	}
	if (ntohl(*bufp++) != 2) {		/* RPC version != 2 */
		nfsdstats.rpcbadfmt++;
		nfsd_reject_version(rqstp);
		goto finish;
	}
	if (ntohl(*bufp++) != 100003) {	/* prog != NFS */
		nfsdstats.rpcbadfmt++;
		rpc_stat = rpc_prog_unavail;
		goto accept_rply;
	}
	rqstp->rq_vers = vers = ntohl(*bufp++);
	if (vers > NFSD_MAXVERS || !(vp = nfsd_version + vers)->sv_procs) {
		rpc_stat = rpc_prog_mismatch;	/* version not supported */
		goto accept_rply;
	}

	rqstp->rq_proc = proc = ntohl(*bufp++);
	if (proc >= vp->sv_nproc || !(pp = vp->sv_procs + proc)->p_func) {
		rpc_stat = rpc_proc_unavail;	/* Procedure not supported. */
		goto accept_rply;
	}

	argp->buf += 5;
	argp->len -= 5;

	/* decode auth data. */
	if (!nfsd_decode_auth(rqstp, argp, &rqstp->rq_cred)) {
		nfsdstats.rpcbadfmt++;
		rpc_stat = rpc_garbage_args;
		goto accept_rply;
	}

	/* Syntactic check complete */
	nfsdstats.rpccnt++;

	/* Validate authentication info. */
	if (!nfsd_validate_auth(rqstp, &auth_stat)) {
		nfsdstats.rpcbadauth++;
		nfsd_reject_auth(rqstp, auth_stat);
		goto finish;
	}

	/* XXX: Ideally, we would want to check here whether this
	 * authentication flavor is acceptable. Unfortunately, we don't
	 * have the exports flags yet, so we have to defer that until
	 * exp_lookup or nfsd_permission.
	 */

accept_rply:
	/* Build the reply header.  */
	statp = nfsd_accept_reply(rqstp, rpc_stat);

	if (rpc_stat != rpc_success)
		return 1;

	/* Check whether we have this call in the cache. */
	switch (nfscache_lookup(rqstp, pp->p_cachetype)) {
	case RC_INTR:
	case RC_DROPIT:
		return 0;
	case RC_REPLY:
		return 1;
	case RC_DOIT:
		/* do it */
	}

	/*
	 * Now call the pp handler.
	 */
	*statp = rpc_stat = pp->p_func(rqstp, argp, resp);

	if (rpc_stat != rpc_success) {
		nfscache_update(rqstp, RC_NOCACHE, NULL);
		resp->len = statp + 1 - resp->base;
		return 1;
	}

	/* Update statistics */
	pp->p_time = jiffies - t0;
	pp->p_count++;

	/* Store reply in cache */
	nfscache_update(rqstp, pp->p_cachetype, statp + 1);

finish:
	nfsd_free_auth(&rqstp->rq_cred);

	return 1;
}


#define DECODE_AUTH(buf, len, flavor, obase, olen)	\
	if (len < 2)					\
		return 0;				\
	flavor = ntohl(*buf++); len--;			\
	olen   = ntohl(*buf++); len--;			\
	if ((olen & 3) || (len < (olen >>= 2)))		\
		return 0;				\
	obase = buf;					\
	buf += olen; len -= olen;

static int
nfsd_decode_auth(svc_req *rqstp, svc_buf *argp, svc_auth *ahp)
{
	u32	*buf = argp->buf;
	int	len = argp->len;

	DECODE_AUTH(buf, len, ahp->ah_crflavor, ahp->ah_crbase, ahp->ah_crlen);
	DECODE_AUTH(buf, len, ahp->ah_vfflavor, ahp->ah_vfbase, ahp->ah_vflen);

	ahp->ah_op = auth_getops(ahp->ah_crflavor);

	if (!ahp->ah_op || !ahp->ah_op->decode(ahp))
		return 0;

	argp->buf = buf;
	argp->len = len;

	return 1;
}

/*
 * Build the reply header. For now, we fill in just a NULL verifier,
 * but we should really call auth->ah_ops->addverf().
 */
u32 *
nfsd_accept_reply(svc_req *rqstp, u32 rpc_stat)
{
	u32 *	bufp = rqstp->rq_res.buf;

	if (rqstp->rq_res.buflen - rqstp->rq_res.len < 5)
		panic("nfsd: reply buffer too small");
	*bufp++ = xdr_one;		/* REPLY */
	*bufp++ = xdr_zero;		/* ACCEPTED */
	*bufp++ = xdr_zero;		/* NULL verf for now */
	*bufp++ = xdr_zero;
	*bufp++ = rpc_stat;
	rqstp->rq_res.len += bufp - rqstp->rq_res.buf;
	rqstp->rq_res.buf = bufp;

	return bufp - 1;		/* return pointer to RPC status */
}

/*
 * Bad RPC version.
 */
void
nfsd_reject_version(svc_req *rqstp)
{
	u32 *	bufp = rqstp->rq_res.buf;

	printk(KERN_INFO "nfsd: reject (rpc version)\n");
	if (rqstp->rq_res.buflen - rqstp->rq_res.len < 5)
		panic("nfsd: reply buffer too small");
	*bufp++ = xdr_one;		/* REPLY */
	*bufp++ = xdr_one;		/* REJECTED */
	*bufp++ = xdr_zero;		/* RPC_MISMATCH */
	*bufp++ = xdr_two;		/* Only RPCv2 supported */
	*bufp++ = xdr_two;
	rqstp->rq_res.len += 5;
	rqstp->rq_res.buf = bufp;
}

void
nfsd_reject_auth(svc_req *rqstp, u32 auth_stat)
{
	u32 *	bufp = rqstp->rq_res.buf;

	printk(KERN_INFO "nfsd: reject (auth error %ld)\n", ntohl(auth_stat));
	if (rqstp->rq_res.buflen - rqstp->rq_res.len < 5)
		panic("nfsd: reply buffer too small");
	*bufp++ = xdr_one;		/* REPLY */
	*bufp++ = xdr_one;		/* REJECTED */
	*bufp++ = xdr_one;		/* AUTH_ERROR */
	*bufp++ = auth_stat;
	rqstp->rq_res.len += 5;
	rqstp->rq_res.buf = bufp;
}

/*
 * This is needed for 1.3.x asm/segment.h
 */
int
bad_user_access_length(void)
{
	panic("bad_user_access_length executed (not cool, dude)");
}
