/*
 * nfsd			The daemon part of nfsd. Still very primitive; may
 *			become slightly more sophisticated.
 *
 * Copyright (C) 1995 Olaf Kirch <okir@monad.swb.de>
 */
#include <linux/unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <rpc/rpc.h>
#include <rpc/pmap_clnt.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#include "nfslib.h"

static int	makesock(int type, int proto);
static int	rpc_register(int program, int version, int port);
static int	rundaemon(int udp, int tcp);
static void	killall(int sig);
static void	usage(void);

#define NFSMAXSERVS	16
static int	child[NFSMAXSERVS];
static int	nchildren = 0;

int
main(int argc, char **argv)
{
	int	count = 1, nfork, i, c, udp, tcp;
	int	foreground = 0;

	xlog_open("nfsd");

	if ((udp = makesock(SOCK_DGRAM, IPPROTO_UDP)) < 0 ||
	    (tcp = makesock(SOCK_STREAM, IPPROTO_TCP)) < 0)
		exit(1);

	while ((c = getopt(argc, argv, "hn")) != EOF) {
		switch(c) {
		case 'n':
			foreground = 1;
			break;
		case 'h':
		default:
			usage();
		}
	}

	if (optind < argc) {
		if ((count = atoi(argv[optind])) > NFSMAXSERVS) {
			/* insane # of servers */
			fprintf(stderr,
				"nfsd: server count too high, using %d\n",
				NFSMAXSERVS);
			count = NFSMAXSERVS;
		}
	}

	if (!rpc_register(100003, 2, 2049))
		xlog(L_FATAL, "can register NFSv2 with rpc.portmap\n");
#ifdef NFS3_SUPPORTED
	if (!rpc_register(100003, 3, 2049))
		xlog(L_FATAL, "can register NFSv3 with rpc.portmap\n");
#endif

	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_DFL);
	setsid();

	nfork = foreground? count - 1 : count;

	for (i = 0; i < nfork; i++) {
		int	pid;

		if ((pid = fork()) < 0) {
			xlog(L_ERROR, "can't fork: %s",
				strerror(errno));
			child[i] = 0;
		} else if (pid) {
			child[i] = pid;
		} else {
			rundaemon(udp, tcp);
		}
	}

	if (foreground) {
		signal(SIGINT, killall);
		signal(SIGTERM, killall);
		nchildren = nfork;
		rundaemon(udp, tcp);
	}

	return 0;
}

static int
makesock(int type, int proto)
{
	struct sockaddr_in	sin;
	int		fd;

	bzero((char *)&sin, sizeof (sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(2049);
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	if ((fd = socket(AF_INET, type, proto)) < 0) {
		perror("can't create socket");
		return fd;
	}
	if (bind(fd, (struct sockaddr*) &sin, sizeof(sin)) == -1) {
		perror("cant bind to port 2049");
		close(fd);
		return -1;
	}
	if (type == SOCK_STREAM) {
		if (listen(fd, 5) < 0) {
			perror("can't listen on port 2049");
			close(fd);
			return -1;
		}
	}

	return fd;
}

static int
rpc_register(int program, int version, int port)
{
	return pmap_unset(program, version)
	    && pmap_set(program, version, IPPROTO_UDP, port)
	    && pmap_set(program, version, IPPROTO_TCP, port);
}

static int
rundaemon(int udp, int tcp)
{
	int	i;

	for (i = 0; i < NR_OPEN; i++)
		if (i != udp && i != tcp)
			close(i);
	xlog_background();

	if (nfssvc(udp, tcp) < 0)
		xlog(L_ERROR, "nfssvc returned: %s", strerror(errno));

	exit(1);
}

static void
killall(int sig)
{
	int	i;

	for (i = 0; i < nchildren; i++)
		if (child[i])
			kill(SIGTERM, child[i]);
	exit(1);
}

static void
usage(void)
{
	fprintf(stderr, "usage: nfsd [-n] [ncopies]\n");
	exit(2);
}
