/*
 * nfsstat.c		Output server-side NFS statistics
 *
 * Copyright (C) 1995, 1996 Olaf Kirch <okir@monad.swb.de>
 */

#define NFSSRVSTAT	"/proc/net/nfssrv"

#include <nfs/stats.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#define MAXNRVALS	32

static unsigned int	nfsv2[18];	/* NFSv2 call counts */
static unsigned int	netcnt;		/* # of received packets */
static unsigned int	netudpcnt;	/* UDP packets */
static unsigned int	nettcpcnt;	/* TCP packets */
static unsigned int	nettcpconn;	/* TCP connections */

static unsigned int	rpccnt;		/* total # of RPC calls */
static unsigned int	rpcbadfmt;	/* bad format */
static unsigned int	rpcbadauth;	/* authentication failed */
static unsigned int	rpcbadclnt;	/* unknown client */

static unsigned int	fhhits;		/* fh cache hits */
static unsigned int	fhmisses;	/* fh cache misses */
static unsigned int	fhstale;	/* stale fhs */
static unsigned int	fhcksum;	/* checksum computations */
static unsigned int	fhbadcksum;	/* bad checksum */

static unsigned int	rchits;		/* repcache hits */
static unsigned int	rcmisses;	/* repcache hits */
static unsigned int	rcnocache;	/* uncached reqs */

static char *	nfsv2name[18] = {
	"null", "getattr", "setattr", "root",   "lookup",  "readlink",
	"read", "wrcache", "write",   "create", "remove",  "rename",
	"link", "symlink", "mkdir",   "rmdir",  "readdir", "fsstat"
};

typedef struct statinfo {
	char		*tag;
	int		nrvals;
	unsigned int *	valptr[MAXNRVALS];
} statinfo;

static int		parse_srv_stat(void);

#define PRNT_NFS	0x0001
#define PRNT_RPC	0x0002
#define PRNT_NET	0x0004
#define PRNT_FH		0x0008
#define PRNT_RC		0x0010
#define PRNT_ALL	0xffff

int
main(int argc, char **argv)
{
	int		opt_all = 0,
			opt_srv = 0,
			opt_clt = 0,
			opt_prt = 0;
	int		c;

	while ((c = getopt(argc, argv, "acno:rsz")) != -1) {
		switch (c) {
		case 'a':
			opt_all = 1;
		case 'c':
			opt_clt = 1;
			break;
		case 'n':
			opt_prt |= PRNT_NFS;
			break;
		case 'o':
			if (!strcmp(optarg, "nfs"))
				opt_prt |= PRNT_NFS;
			else if (!strcmp(optarg, "rpc"))
				opt_prt |= PRNT_RPC;
			else if (!strcmp(optarg, "net"))
				opt_prt |= PRNT_NET;
			else if (!strcmp(optarg, "fh"))
				opt_prt |= PRNT_FH;
			else if (!strcmp(optarg, "rc"))
				opt_prt |= PRNT_RC;
			else {
				fprintf(stderr, "nfsstat: unknown category: "
						"%s\n", optarg);
				return 2;
			}
		case 'r':
			opt_prt |= PRNT_RPC;
			break;
		case 's':
			opt_srv = 1;
			break;
		case 'z':
			fprintf(stderr, "nfsstat: zeroing of nfs statistics "
					"not yet supported\n");
			return 2;
		}
	}

	if (opt_clt) {
		fprintf(stderr, "nfsstat: client side NFS stats not "
				"yet supported -- ignoring\n");
	}

	if (opt_all) {
		opt_srv = opt_clt = 1;
		opt_prt = PRNT_ALL;
	}
	if (!(opt_srv + opt_clt))
		opt_srv = opt_clt = 1;
	if (!opt_prt)
		opt_prt = PRNT_NFS + PRNT_RPC;

	if (!(parse_srv_stat()))
		return 2;

	if (opt_prt & PRNT_NET) {
		printf("Server net:\n"
		       "packets    udp        tcp        tcpconn\n"
		       "%-8d   %-8d   %-8d   %-8d\n",
		       netcnt, netudpcnt, nettcpcnt, nettcpconn);
		printf("\n");
	}

	if (opt_prt & PRNT_RPC) {
		unsigned int	rpcbadcalls;

		rpcbadcalls = rpcbadauth + rpcbadclnt + rpcbadfmt;
		printf("Server rpc:\n"
		       "calls      badcalls   badauth    badclnt    xdrcall\n"
		       "%-8d   %-8d   %-8d   %-8d   %-8d\n",
		       rpccnt, rpcbadcalls, rpcbadauth, rpcbadclnt, rpcbadfmt);
		printf("\n");
	}

	if (opt_prt & PRNT_FH) {
		printf("Server file handle cache:\n"
		       "hits       misses     stalefh    cksum      badcksum\n"
		       "%-8d   %-8d   %-8d   %-8d   %-8d\n",
		       fhhits, fhmisses, fhstale, fhcksum, fhbadcksum);
		printf("\n");
	}

	if (opt_prt & PRNT_RC) {
		printf("Server reply cache:\n"
		       "hits       misses     nocache\n"
		       "%-8d   %-8d   %-8d\n",
		       rchits, rcmisses, rcnocache);
		printf("\n");
	}

	if (opt_prt & PRNT_NFS) {
		unsigned int	total;
		int		i, j;

		printf("Server nfs v2:\n");
		for (i = 0, total = 0; i < 18; i++)
			total += nfsv2[i];
		for (i = 0; i < 18; i += 6) {
			for (j = 0; j < 6; j++)
				printf("%-11s", nfsv2name[i+j]);
			printf("\n");
			for (j = 0; j < 6; j++)
				printf("%-6d %2d%% ",
					nfsv2[i+j], 100 * nfsv2[i+j] / total);
			printf("\n");
		}
		printf("\n");
	}

	return 0;
}

static statinfo		servinfo[] = {
    {	"net",	4,	{ &netcnt,
    			  &netudpcnt,
			  &nettcpcnt,
			  &nettcpconn } },
    {	"rpc",	4,	{ &rpccnt,
			  &rpcbadfmt,
			  &rpcbadauth,
			  &rpcbadclnt } },
    {	"fh",	5,	{ &fhhits,
			  &fhmisses,
			  &fhstale,
			  &fhcksum,
			  &fhbadcksum } },
    {	"rc",	3,	{ &rchits,
			  &rcmisses,
			  &rcnocache } },
    {	"nfsv2",18,	{ nfsv2 + 0,
    			  nfsv2 + 1,
    			  nfsv2 + 2,
    			  nfsv2 + 3,
    			  nfsv2 + 4,
    			  nfsv2 + 5,
    			  nfsv2 + 6,
    			  nfsv2 + 7,
    			  nfsv2 + 8,
    			  nfsv2 + 9,
    			  nfsv2 + 10,
    			  nfsv2 + 11,
    			  nfsv2 + 12,
    			  nfsv2 + 13,
    			  nfsv2 + 14,
    			  nfsv2 + 15,
    			  nfsv2 + 16,
    			  nfsv2 + 17 } },

    {	NULL,	0,	{ 0 } }
};

static int
parse_srv_stat(void)
{
	char	buffer[4096], *line, *next;
	int	fd, len;

	if ((fd = open(NFSSRVSTAT, O_RDONLY)) < 0) {
		perror("nfsstat: can't open " NFSSRVSTAT "\n");
		return 0;
	}

	if ((len = read(fd, buffer, sizeof(buffer))) < 0) {
		perror("nfsstat: can't read from " NFSSRVSTAT);
		close(fd);
		return 0;
	}

	close(fd);

	for (line = buffer; line; line = next) {
		statinfo	*ip;
		char		*sp;
		int		i, cnt;

		if ((next = strchr(line, '\n')) != NULL)
			*next++ = '\0';
		if (!(sp = strtok(line, " \t")))
			continue;
		for (ip = servinfo; ip->tag; ip++) {
			if (!strcmp(sp, ip->tag))
				break;
		}
		if (!ip->tag)
			continue;
		cnt = ip->nrvals;

		for (i = 0; i < cnt; i++) {
			if (!(sp = strtok(NULL, " \t")))
				break;
			*(ip->valptr[i]) = atoi(sp);
		}
	}

	return 1;
}
